@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 /* Document interface */
 
@@ -584,6 +584,21 @@ FUN(Document_recognize)(JNIEnv *env, jclass cls, jstring jmagic)
 	return recognized;
 }
 
+JNIEXPORT jboolean JNICALL
+FUN(Document_supportsAccelerator)(JNIEnv *env, jobject self)
+{
+	fz_context *ctx = get_context(env);
+	fz_document *doc = from_Document(env, self);
+	jboolean support = JNI_FALSE;
+
+	fz_try(ctx)
+		support = fz_document_supports_accelerator(ctx, doc);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	return support;
+}
+
 JNIEXPORT void JNICALL
 FUN(Document_saveAccelerator)(JNIEnv *env, jobject self, jstring jfilename)
 {
@@ -814,6 +829,38 @@ FUN(Document_getMetaData)(JNIEnv *env, jobject self, jstring jkey)
 	return (*env)->NewStringUTF(env, info);
 }
 
+JNIEXPORT void JNICALL
+FUN(Document_setMetaData)(JNIEnv *env, jobject self, jstring jkey, jstring jvalue)
+{
+	fz_context *ctx = get_context(env);
+	fz_document *doc = from_Document(env, self);
+	const char *key = NULL;
+	const char *value = NULL;
+
+	if (!ctx || !doc) return;
+	if (!jkey) jni_throw_arg_void(env, "key must not be null");
+	if (!jvalue) jni_throw_arg_void(env, "value must not be null");
+
+	key = (*env)->GetStringUTFChars(env, jkey, NULL);
+	value = (*env)->GetStringUTFChars(env, jvalue, NULL);
+	if (!key || !value)
+	{
+		if (key)
+			(*env)->ReleaseStringUTFChars(env, jkey, key);
+		return;
+	}
+
+	fz_try(ctx)
+		fz_set_metadata(ctx, doc, key, value);
+	fz_always(ctx)
+	{
+		(*env)->ReleaseStringUTFChars(env, jkey, key);
+		(*env)->ReleaseStringUTFChars(env, jvalue, value);
+	}
+	fz_catch(ctx)
+		jni_rethrow_void(env, ctx);
+}
+
 JNIEXPORT jboolean JNICALL
 FUN(Document_isUnencryptedPDF)(JNIEnv *env, jobject self)
 {
@@ -863,6 +910,40 @@ FUN(Document_loadOutline)(JNIEnv *env, jobject self)
 	return joutline;
 }
 
+JNIEXPORT jobject JNICALL
+FUN(Document_outlineIterator)(JNIEnv *env, jobject self)
+{
+	fz_context *ctx = get_context(env);
+	fz_document *doc = from_Document(env, self);
+	fz_outline_iterator *iterator = NULL;
+	jobject jiterator = NULL;
+
+	if (!ctx || !doc) return NULL;
+
+	fz_var(iterator);
+
+	fz_try(ctx)
+	{
+		iterator = fz_new_outline_iterator(ctx, doc);
+		if (iterator)
+		{
+			jiterator = to_OutlineIterator_safe(ctx, env, iterator);
+			if (!jiterator || (*env)->ExceptionCheck(env))
+				fz_throw(ctx, FZ_ERROR_GENERIC, "outlineIterator failed");
+			iterator = NULL;
+		}
+	}
+	fz_always(ctx)
+		fz_drop_outline_iterator(ctx, iterator);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	if ((*env)->ExceptionCheck(env))
+		return NULL;
+
+	return jiterator;
+}
+
 JNIEXPORT jlong JNICALL
 FUN(Document_makeBookmark)(JNIEnv *env, jobject self, jint chapter, jint page)
 {
@@ -940,22 +1021,79 @@ FUN(Document_search)(JNIEnv *env, jobject self, jint chapter, jint page, jstring
 {
 	fz_context *ctx = get_context(env);
 	fz_document *doc = from_Document(env, self);
-	fz_quad hits[256];
+	fz_quad hits[500];
+	int marks[500];
 	const char *needle = NULL;
 	int n = 0;
 
-	if (!ctx || !page) return NULL;
-	if (!jneedle) jni_throw_arg(env, "needle must not be null");
+	if (!ctx) return NULL;
+	if (!jneedle) return NULL;
 
 	needle = (*env)->GetStringUTFChars(env, jneedle, NULL);
 	if (!needle) return 0;
 
 	fz_try(ctx)
-		n = fz_search_chapter_page_number(ctx, doc, chapter, page, needle, hits, nelem(hits));
+		n = fz_search_chapter_page_number(ctx, doc, chapter, page, needle, marks, hits, nelem(hits));
 	fz_always(ctx)
 		(*env)->ReleaseStringUTFChars(env, jneedle, needle);
 	fz_catch(ctx)
 		jni_rethrow(env, ctx);
 
-	return to_QuadArray_safe(ctx, env, hits, n);
+	return to_SearchHits_safe(ctx, env, marks, hits, n);
+}
+
+JNIEXPORT jobject JNICALL
+FUN(Document_resolveLinkDestination)(JNIEnv *env, jobject self, jstring juri)
+{
+	fz_context *ctx = get_context(env);
+	fz_document *doc = from_Document(env, self);
+	const char *uri = "";
+	fz_link_dest dest;
+	jobject jdestination;
+
+	if (!ctx || !doc) return NULL;
+
+	if (juri)
+	{
+		uri = (*env)->GetStringUTFChars(env, juri, NULL);
+		if (!uri)
+			return NULL;
+	}
+
+	fz_try(ctx)
+		dest = fz_resolve_link_dest(ctx, doc, uri);
+	fz_always(ctx)
+		if (juri)
+			(*env)->ReleaseStringUTFChars(env, juri, uri);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	jdestination = (*env)->NewObject(env, cls_LinkDestination, mid_LinkDestination_init,
+		dest.loc.chapter, dest.loc.page, dest.type, dest.x, dest.y, dest.w, dest.h, dest.zoom);
+	if (!jdestination || (*env)->ExceptionCheck(env))
+		return NULL;
+
+	return jdestination;
+}
+
+JNIEXPORT jstring JNICALL
+FUN(Document_formatLinkURI)(JNIEnv *env, jobject self, jobject jdest)
+{
+	fz_context *ctx = get_context(env);
+	fz_document *doc = from_Document(env, self);
+	fz_link_dest dest = from_LinkDestination(env, jdest);
+	char *uri = NULL;
+	jobject juri;
+
+	fz_try(ctx)
+		uri = fz_format_link_uri(ctx, doc, dest);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	juri = (*env)->NewStringUTF(env, uri);
+	fz_free(ctx, uri);
+	if (juri == NULL || (*env)->ExceptionCheck(env))
+		return NULL;
+
+	return juri;
 }
