@@ -25,26 +25,6 @@
 #define MAX(a,b) ((a) > (b) ? (a) : (b))
 #endif
 
-static int
-fz_mkdir(char *path)
-{
-#ifdef _WIN32
-	int ret;
-	wchar_t *wpath = fz_wchar_from_utf8(path);
-
-	if (wpath == NULL)
-		return -1;
-
-	ret = _wmkdir(wpath);
-
-	free(wpath);
-
-	return ret;
-#else
-	return mkdir(path, S_IRWXU | S_IRWXG | S_IRWXO);
-#endif
-}
-
 static int create_accel_path(fz_context *ctx, char outname[], size_t len, int create, const char *absname, ...)
 {
 	va_list args;
@@ -458,6 +438,7 @@ void pdfapp_open_progressive(pdfapp_t *app, char *filename, int reload, int kbps
 				{
 					if (fz_caught(ctx) == FZ_ERROR_TRYLATER)
 					{
+						fz_ignore_error(ctx);
 						pdfapp_sleep(100);
 						continue;
 					}
@@ -480,6 +461,7 @@ void pdfapp_open_progressive(pdfapp_t *app, char *filename, int reload, int kbps
 				{
 					if (fz_caught(ctx) == FZ_ERROR_TRYLATER)
 					{
+						fz_ignore_error(ctx);
 						pdfapp_sleep(100);
 						continue;
 					}
@@ -578,6 +560,7 @@ void pdfapp_open_progressive(pdfapp_t *app, char *filename, int reload, int kbps
 			{
 				if (fz_caught(ctx) == FZ_ERROR_TRYLATER)
 				{
+					fz_ignore_error(ctx);
 					continue;
 				}
 				fz_rethrow(ctx);
@@ -594,9 +577,15 @@ void pdfapp_open_progressive(pdfapp_t *app, char *filename, int reload, int kbps
 			{
 				app->outline = NULL;
 				if (fz_caught(ctx) == FZ_ERROR_TRYLATER)
+				{
+					fz_ignore_error(ctx);
 					app->outline_deferred = PDFAPP_OUTLINE_DEFERRED;
+				}
 				else
+				{
+					fz_report_error(ctx);
 					pdfapp_warn(app, "Failed to load outline.");
+				}
 			}
 			break;
 		}
@@ -845,9 +834,15 @@ static void pdfapp_loadpage(pdfapp_t *app, int no_cache)
 	fz_catch(app->ctx)
 	{
 		if (fz_caught(app->ctx) == FZ_ERROR_TRYLATER)
+		{
+			fz_ignore_error(app->ctx);
 			app->incomplete = 1;
+		}
 		else
+		{
+			fz_report_error(app->ctx);
 			pdfapp_warn(app, "Failed to load page.");
+		}
 		return;
 	}
 
@@ -886,9 +881,15 @@ static void pdfapp_loadpage(pdfapp_t *app, int no_cache)
 		fz_catch(app->ctx)
 		{
 			if (fz_caught(app->ctx) == FZ_ERROR_TRYLATER)
+			{
+				fz_ignore_error(app->ctx);
 				app->incomplete = 1;
+			}
 			else
+			{
+				fz_report_error(app->ctx);
 				pdfapp_warn(app, "Failed to load page.");
+			}
 			errored = 1;
 		}
 	}
@@ -926,9 +927,15 @@ static void pdfapp_loadpage(pdfapp_t *app, int no_cache)
 	fz_catch(app->ctx)
 	{
 		if (fz_caught(app->ctx) == FZ_ERROR_TRYLATER)
+		{
+			fz_ignore_error(app->ctx);
 			app->incomplete = 1;
+		}
 		else
+		{
+			fz_report_error(app->ctx);
 			pdfapp_warn(app, "Failed to load page.");
+		}
 		errored = 1;
 	}
 
@@ -1000,8 +1007,8 @@ static void pdfapp_showpage(pdfapp_t *app, int loadpage, int drawpage, int repai
 			mediabox = fz_bound_page(app->ctx, app->page);
 		fz_catch(app->ctx)
 		{
-			if (fz_caught(app->ctx) != FZ_ERROR_TRYLATER)
-				fz_rethrow(app->ctx);
+			fz_rethrow_unless(app->ctx, FZ_ERROR_TRYLATER);
+			fz_ignore_error(app->ctx);
 			mediabox = fz_make_rect(0, 0, 100, 100);
 			app->incomplete = 1;
 		}
