@@ -62,6 +62,8 @@ fz_disable_device(fz_context *ctx, fz_device *dev)
 	dev->end_layer = NULL;
 	dev->begin_structure = NULL;
 	dev->end_structure = NULL;
+	dev->begin_metatext = NULL;
+	dev->end_metatext = NULL;
 }
 
 void
@@ -141,7 +143,7 @@ pop_clip_stack(fz_context *ctx, fz_device *dev, int type)
 	if (dev->container_len == 0 || dev->container[dev->container_len-1].type != type)
 	{
 		fz_disable_device(ctx, dev);
-		fz_throw(ctx, FZ_ERROR_GENERIC, "device calls unbalanced");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "device calls unbalanced");
 	}
 	dev->container_len--;
 }
@@ -152,7 +154,7 @@ pop_push_clip_stack(fz_context *ctx, fz_device *dev, int pop_type, int push_type
 	if (dev->container_len == 0 || dev->container[dev->container_len-1].type != pop_type)
 	{
 		fz_disable_device(ctx, dev);
-		fz_throw(ctx, FZ_ERROR_GENERIC, "device calls unbalanced");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "device calls unbalanced");
 	}
 	dev->container[dev->container_len-1].type = push_type;
 }
@@ -348,7 +350,7 @@ void
 fz_fill_image(fz_context *ctx, fz_device *dev, fz_image *image, fz_matrix ctm, float alpha, fz_color_params color_params)
 {
 	if (image->colorspace == NULL)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "argument to fill image must be a color image");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "argument to fill image must be a color image");
 	if (dev->fill_image)
 	{
 		fz_try(ctx)
@@ -414,14 +416,14 @@ fz_begin_mask(fz_context *ctx, fz_device *dev, fz_rect area, int luminosity, fz_
 }
 
 void
-fz_end_mask(fz_context *ctx, fz_device *dev)
+fz_end_mask_tr(fz_context *ctx, fz_device *dev, fz_function *fn)
 {
 	pop_push_clip_stack(ctx, dev, fz_device_container_stack_is_mask, fz_device_container_stack_is_clip);
 
 	if (dev->end_mask)
 	{
 		fz_try(ctx)
-			dev->end_mask(ctx, dev);
+			dev->end_mask(ctx, dev, fn);
 		fz_catch(ctx)
 		{
 			fz_disable_device(ctx, dev);
@@ -430,6 +432,12 @@ fz_end_mask(fz_context *ctx, fz_device *dev)
 	}
 }
 
+void
+fz_end_mask(fz_context *ctx, fz_device *dev)
+{
+	fz_end_mask_tr(ctx, dev, NULL);
+}
+
 void
 fz_begin_group(fz_context *ctx, fz_device *dev, fz_rect area, fz_colorspace *cs, int isolated, int knockout, int blendmode, float alpha)
 {
@@ -570,12 +578,12 @@ void fz_end_layer(fz_context *ctx, fz_device *dev)
 	}
 }
 
-void fz_begin_structure(fz_context *ctx, fz_device *dev, fz_structure str, const char *raw, int uid)
+void fz_begin_structure(fz_context *ctx, fz_device *dev, fz_structure str, const char *raw, int idx)
 {
 	if (dev->begin_structure)
 	{
 		fz_try(ctx)
-			dev->begin_structure(ctx, dev, str, raw, uid);
+			dev->begin_structure(ctx, dev, str, raw, idx);
 		fz_catch(ctx)
 		{
 			fz_disable_device(ctx, dev);
@@ -835,3 +843,64 @@ fz_structure_from_string(const char *str)
 	if (!strcmp(str, "Form")) return FZ_STRUCTURE_FORM;
 	return FZ_STRUCTURE_INVALID;
 }
+
+fz_function *
+fz_keep_function(fz_context *ctx, fz_function *func)
+{
+	return fz_keep_storable(ctx, &func->storable);
+}
+
+void
+fz_drop_function(fz_context *ctx, fz_function *func)
+{
+	fz_drop_storable(ctx, &func->storable);
+}
+
+size_t
+fz_function_size(fz_context *ctx, fz_function *func)
+{
+	return (func ? func->size : 0);
+}
+
+void
+fz_eval_function(fz_context *ctx, fz_function *func, const float *in, int inlen, float *out, int outlen)
+{
+	float fakein[FZ_FUNCTION_MAX_M];
+	float fakeout[FZ_FUNCTION_MAX_N];
+	int i;
+
+	if (inlen < func->m)
+	{
+		for (i = 0; i < inlen; ++i)
+			fakein[i] = in[i];
+		for (; i < func->m; ++i)
+			fakein[i] = 0;
+		in = fakein;
+	}
+
+	if (outlen < func->n)
+	{
+		func->eval(ctx, func, in, fakeout);
+		for (i = 0; i < outlen; ++i)
+			out[i] = fakeout[i];
+	}
+	else
+	{
+		func->eval(ctx, func, in, out);
+		for (i = func->n; i < outlen; ++i)
+			out[i] = 0;
+	}
+}
+
+fz_function *
+fz_new_function_of_size(fz_context *ctx, int size, size_t size2, int m, int n, fz_function_eval_fn *eval, fz_store_drop_fn *drop)
+{
+	fz_function *fn = fz_calloc(ctx, 1, size);
+
+	FZ_INIT_STORABLE(fn, 1, drop);
+	fn->eval = eval;
+	fn->m = m;
+	fn->n = n;
+
+	return fn;
+}
