@@ -103,9 +103,9 @@ static const unsigned char *
 pnm_read_signature(fz_context *ctx, const unsigned char *p, const unsigned char *e, char *signature)
 {
 	if (e - p < 2)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse magic number in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "cannot parse magic number in pnm image");
 	if (p[0] != 'P' || ((p[1] < '1' || p[1] > '7') && p[1] != 'F' && p[1] != 'f'))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "expected signature in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "expected signature in pnm image");
 
 	signature[0] = *p++;
 	signature[1] = *p++;
@@ -116,7 +116,7 @@ static const unsigned char *
 pnm_read_until_eol(fz_context *ctx, const unsigned char *p, const unsigned char *e, int acceptCR)
 {
 	if (e - p < 1)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse line in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "cannot parse line in pnm image");
 
 	while (p < e && ((acceptCR && *p != '\r' && *p != '\n') || (!acceptCR && *p != '\n')))
 		p++;
@@ -128,9 +128,9 @@ static const unsigned char *
 pnm_read_eol(fz_context *ctx, const unsigned char *p, const unsigned char *e, int acceptCR)
 {
 	if (e - p < 1)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse end of line in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "cannot parse end of line in pnm image");
 	if ((acceptCR && *p != '\r' && *p != '\n') || (!acceptCR && *p != '\n'))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "expected end of line in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "expected end of line in pnm image");
 
 	/* CR, CRLF or LF depending on acceptCR. */
 	if (acceptCR && *p == '\r')
@@ -145,9 +145,9 @@ static const unsigned char *
 pnm_read_whites(fz_context *ctx, const unsigned char *p, const unsigned char *e, int required)
 {
 	if (required && e - p < 1)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse whitespaces in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "cannot parse whitespaces in pnm image");
 	if (required && !iswhite(*p))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "expected whitespaces in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "expected whitespaces in pnm image");
 
 	while (p < e && iswhite(*p))
 		p++;
@@ -159,9 +159,9 @@ static const unsigned char *
 pnm_read_white_or_eol(fz_context *ctx, const unsigned char *p, const unsigned char *e)
 {
 	if (e - p < 1)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse whitespace/eol in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "cannot parse whitespace/eol in pnm image");
 	if (!iswhiteeol(*p))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "expected whitespace/eol in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "expected whitespace/eol in pnm image");
 
 	return ++p;
 }
@@ -170,9 +170,9 @@ static const unsigned char *
 pnm_read_whites_and_eols(fz_context *ctx, const unsigned char *p, const unsigned char *e, int required)
 {
 	if (required && e - p < 1)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse whitespaces/eols in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "cannot parse whitespaces/eols in pnm image");
 	if (required && !iswhiteeol(*p))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "expected whitespaces/eols in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "expected whitespaces/eols in pnm image");
 
 	while (p < e && iswhiteeol(*p))
 		p++;
@@ -184,7 +184,7 @@ static const unsigned char *
 pnm_read_comment(fz_context *ctx, const unsigned char *p, const unsigned char *e, int acceptCR)
 {
 	if (e - p < 1)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse line in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "cannot parse line in pnm image");
 
 	if (*p != '#')
 		return p;
@@ -196,7 +196,7 @@ static const unsigned char *
 pnm_read_comments(fz_context *ctx, const unsigned char *p, const unsigned char *e, int acceptCR)
 {
 	if (e - p < 1)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse comment in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "cannot parse comment in pnm image");
 
 	while (p < e && *p == '#')
 	{
@@ -211,9 +211,9 @@ static const unsigned char *
 pnm_read_digit(fz_context *ctx, const unsigned char *p, const unsigned char *e, int *number)
 {
 	if (e - p < 1)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse digit in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "cannot parse digit in pnm image");
 	if (*p < '0' || *p > '1')
-		fz_throw(ctx, FZ_ERROR_GENERIC, "expected digit in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "expected digit in pnm image");
 
 	if (number)
 		*number = *p - '0';
@@ -226,9 +226,9 @@ static const unsigned char *
 pnm_read_int(fz_context *ctx, const unsigned char *p, const unsigned char *e, int *number)
 {
 	if (e - p < 1)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse integer in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "cannot parse integer in pnm image");
 	if (*p < '0' || *p > '9')
-		fz_throw(ctx, FZ_ERROR_GENERIC, "expected integer in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "expected integer in pnm image");
 
 	while (p < e && *p >= '0' && *p <= '9')
 	{
@@ -248,10 +248,10 @@ pnm_read_real(fz_context *ctx, const unsigned char *p, const unsigned char *e, f
 	size_t len;
 
 	if (e - p < 1)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse real in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "cannot parse real in pnm image");
 
 	if (*p != '+' && *p != '-' && (*p < '0' || *p > '9'))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "expected numeric field in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "expected numeric field in pnm image");
 
 	while (p < e && (*p == '+' || *p == '-' || *p == '.' || (*p >= '0' && *p <= '9')))
 		p++;
@@ -292,7 +292,7 @@ pnm_read_tupletype(fz_context *ctx, const unsigned char *p, const unsigned char
 	int i, len;
 
 	if (e - p < 1)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse tuple type in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "cannot parse tuple type in pnm image");
 
 	s = p;
 	while (p < e && !iswhiteeol(*p))
@@ -306,7 +306,7 @@ pnm_read_tupletype(fz_context *ctx, const unsigned char *p, const unsigned char
 			return p;
 		}
 
-	fz_throw(ctx, FZ_ERROR_GENERIC, "unknown tuple type in pnm image");
+	fz_throw(ctx, FZ_ERROR_FORMAT, "unknown tuple type in pnm image");
 }
 
 static const unsigned char *
@@ -325,7 +325,7 @@ pnm_read_token(fz_context *ctx, const unsigned char *p, const unsigned char *e,
 	int i, len;
 
 	if (e - p < 1)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot parse header token in pnm image");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "cannot parse header token in pnm image");
 
 	s = p;
 	while (p < e && !iswhiteeol(*p))
@@ -339,7 +339,7 @@ pnm_read_token(fz_context *ctx, const unsigned char *p, const unsigned char *e,
 			return p;
 		}
 
-	fz_throw(ctx, FZ_ERROR_GENERIC, "unknown header token in pnm image");
+	fz_throw(ctx, FZ_ERROR_FORMAT, "unknown header token in pnm image");
 }
 
 static int
@@ -381,16 +381,16 @@ pnm_ascii_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p,
 	}
 
 	if (pnm->maxval <= 0 || pnm->maxval >= 65536)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "maximum sample value of out range in pnm image: %d", pnm->maxval);
+		fz_throw(ctx, FZ_ERROR_FORMAT, "maximum sample value of out range in pnm image: %d", pnm->maxval);
 
 	pnm->bitdepth = bitdepth_from_maxval(pnm->maxval);
 
 	if (pnm->height <= 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "image height must be > 0");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "image height must be > 0");
 	if (pnm->width <= 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "image width must be > 0");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "image width must be > 0");
 	if ((unsigned int)pnm->height > UINT_MAX / pnm->width / fz_colorspace_n(ctx, pnm->cs) / (pnm->bitdepth / 8 + 1))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "image too large");
+		fz_throw(ctx, FZ_ERROR_LIMIT, "image too large");
 
 	if (onlymeta)
 	{
@@ -477,6 +477,10 @@ pnm_binary_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p,
 {
 	fz_pixmap *img = NULL;
 	size_t span;
+	int n;
+
+	n = fz_colorspace_n(ctx, pnm->cs);
+	assert(n >= 1 && n <= 3);
 
 	pnm->width = 0;
 	p = pnm_read_comments(ctx, p, e, 1);
@@ -505,33 +509,33 @@ pnm_binary_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p,
 	}
 
 	if (pnm->maxval <= 0 || pnm->maxval >= 65536)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "maximum sample value of out range in pnm image: %d", pnm->maxval);
+		fz_throw(ctx, FZ_ERROR_FORMAT, "maximum sample value of out range in pnm image: %d", pnm->maxval);
 
 	pnm->bitdepth = bitdepth_from_maxval(pnm->maxval);
 
 	if (pnm->height <= 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "image height must be > 0");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "image height must be > 0");
 	if (pnm->width <= 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "image width must be > 0");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "image width must be > 0");
 	if (pnm->bitdepth == 1)
 	{
 		/* Overly sensitive test, but we can live with it. */
-		if ((size_t)pnm->width > SIZE_MAX / (unsigned int)fz_colorspace_n(ctx, pnm->cs))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "image row too large");
-		span = ((size_t)fz_colorspace_n(ctx, pnm->cs) * pnm->width + 7)/8;
+		if ((size_t)pnm->width > SIZE_MAX / (unsigned int)n)
+			fz_throw(ctx, FZ_ERROR_LIMIT, "image row too large");
+		span = ((size_t)n * pnm->width + 7)/8;
 	}
 	else
 	{
 		size_t bytes_per_sample = (pnm->bitdepth-1)/8 + 1;
-		span = (size_t)fz_colorspace_n(ctx, pnm->cs) * bytes_per_sample;
+		span = (size_t)n * bytes_per_sample;
 		if ((size_t)pnm->width > SIZE_MAX / span)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "image row too large");
+			fz_throw(ctx, FZ_ERROR_LIMIT, "image row too large");
 		span = (size_t)pnm->width * span;
 	}
 	if ((size_t)pnm->height > SIZE_MAX / span)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "image too large");
+		fz_throw(ctx, FZ_ERROR_LIMIT, "image too large");
 	if (e - p < 0 || ((size_t)(e - p)) < span * (size_t)pnm->height)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "insufficient data");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "insufficient data");
 
 	if (onlymeta)
 	{
@@ -618,7 +622,7 @@ pam_binary_read_header(fz_context *ctx, struct info *pnm, const unsigned char *p
 			p = pnm_read_token(ctx, p, eol, &token);
 
 			if (seen[token - 1])
-				fz_throw(ctx, FZ_ERROR_GENERIC, "token occurs multiple times in pnm image");
+				fz_throw(ctx, FZ_ERROR_FORMAT, "token occurs multiple times in pnm image");
 			seen[token - 1] = 1;
 
 			if (token != TOKEN_ENDHDR)
@@ -667,7 +671,7 @@ pam_binary_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p,
 		case 4: pnm->tupletype = PAM_CMYK; break;
 		case 5: pnm->tupletype = PAM_CMYKA; break;
 		default:
-			fz_throw(ctx, FZ_ERROR_GENERIC, "cannot guess tuple type based on depth in pnm image");
+			fz_throw(ctx, FZ_ERROR_FORMAT, "cannot guess tuple type based on depth in pnm image");
 		}
 
 	if (pnm->tupletype == PAM_BW && pnm->maxval > 1)
@@ -709,22 +713,22 @@ pam_binary_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p,
 		pnm->cs = fz_device_cmyk(ctx);
 		break;
 	default:
-		fz_throw(ctx, FZ_ERROR_GENERIC, "unsupported tuple type");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "unsupported tuple type");
 	}
 
 	if (pnm->depth != fz_colorspace_n(ctx, pnm->cs) + pnm->alpha)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "depth out of tuple type range");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "depth out of tuple type range");
 	if (pnm->maxval < minval || pnm->maxval > maxval)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "maxval out of range");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "maxval out of range");
 
 	pnm->bitdepth = bitdepth_from_maxval(pnm->maxval);
 
 	if (pnm->height <= 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "image height must be > 0");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "image height must be > 0");
 	if (pnm->width <= 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "image width must be > 0");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "image width must be > 0");
 	if ((unsigned int)pnm->height > UINT_MAX / pnm->width / fz_colorspace_n(ctx, pnm->cs) / (pnm->bitdepth / 8 + 1))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "image too large");
+		fz_throw(ctx, FZ_ERROR_LIMIT, "image too large");
 
 	if (onlymeta)
 	{
@@ -748,9 +752,9 @@ pam_binary_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p,
 				packed = 1;
 		}
 		if (packed && (e < p || (size_t)(e - p) < size / 8))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "truncated packed image");
+			fz_throw(ctx, FZ_ERROR_FORMAT, "truncated packed image");
 		if (!packed && (e < p || (size_t)(e - p) < size * (pnm->maxval < 256 ? 1 : 2)))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "truncated image");
+			fz_throw(ctx, FZ_ERROR_FORMAT, "truncated image");
 
 		if (pnm->maxval == 255)
 			p += size;
@@ -791,9 +795,9 @@ pam_binary_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p,
 					packed = 1;
 			}
 			if (packed && (e < p || (size_t)(e - p) < size / 8))
-				fz_throw(ctx, FZ_ERROR_GENERIC, "truncated packed image");
+				fz_throw(ctx, FZ_ERROR_FORMAT, "truncated packed image");
 			if (!packed && (e < p || (size_t)(e - p) < size * (pnm->maxval < 256 ? 1 : 2)))
-				fz_throw(ctx, FZ_ERROR_GENERIC, "truncated image");
+				fz_throw(ctx, FZ_ERROR_FORMAT, "truncated image");
 
 			if (pnm->maxval == 255)
 				memcpy(dp, p, size);
@@ -890,11 +894,11 @@ pfm_binary_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p,
 	pnm->cs = rgb ? fz_device_rgb(ctx) : fz_device_gray(ctx);
 
 	if (pnm->height <= 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "image height must be > 0");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "image height must be > 0");
 	if (pnm->width <= 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "image width must be > 0");
+		fz_throw(ctx, FZ_ERROR_FORMAT, "image width must be > 0");
 	if ((unsigned int)pnm->height > UINT_MAX / pnm->width / fz_colorspace_n(ctx, pnm->cs) / (pnm->bitdepth / 8 + 1))
-		fz_throw(ctx, FZ_ERROR_GENERIC, "image too large");
+		fz_throw(ctx, FZ_ERROR_LIMIT, "image too large");
 
 	if (onlymeta)
 	{
@@ -904,7 +908,7 @@ pfm_binary_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p,
 		size_t size = w * h * n * sizeof(float);
 
 		if (e < p || (size_t)(e - p) < size)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "truncated image");
+			fz_throw(ctx, FZ_ERROR_FORMAT, "truncated image");
 
 		p += size;
 	}
@@ -919,7 +923,7 @@ pfm_binary_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p,
 		int x, y, k;
 
 		if (e < p || (size_t)(e - p) < size)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "truncated image");
+			fz_throw(ctx, FZ_ERROR_FORMAT, "truncated image");
 
 		sample = samples = fz_malloc(ctx, size);
 		fz_try(ctx)
@@ -1007,7 +1011,7 @@ pnm_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p, size_t
 		else if (!strcmp(signature, "PF"))
 			pix = pfm_binary_read_image(ctx, pnm, p, e, subonlymeta, 1, &p);
 		else
-			fz_throw(ctx, FZ_ERROR_GENERIC, "unsupported portable anymap signature (0x%02x, 0x%02x)", signature[0], signature[1]);
+			fz_throw(ctx, FZ_ERROR_FORMAT, "unsupported portable anymap signature (0x%02x, 0x%02x)", signature[0], signature[1]);
 
 		p = pnm_read_whites_and_eols(ctx, p, e, 0);
 
@@ -1018,7 +1022,7 @@ pnm_read_image(fz_context *ctx, struct info *pnm, const unsigned char *p, size_t
 	}
 
 	if (p >= e && subimage >= 0)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "subimage count out of range");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "subimage count out of range");
 
 	return pix;
 }
