@@ -158,6 +158,8 @@ fz_document_writer *fz_new_pkm_pixmap_writer(fz_context *ctx, const char *path,
 
 static int is_extension(const char *a, const char *ext)
 {
+	if (!a)
+		return 0;
 	if (a[0] == '.')
 		++a;
 	return !fz_strcasecmp(a, ext);
@@ -244,21 +246,27 @@ fz_new_document_writer(fz_context *ctx, const char *path, const char *explicit_f
 		else
 			format = NULL;
 	}
-	fz_throw(ctx, FZ_ERROR_GENERIC, "cannot detect document format");
+	fz_throw(ctx, FZ_ERROR_ARGUMENT, "cannot detect document format");
 }
 
 fz_document_writer *
 fz_new_document_writer_with_output(fz_context *ctx, fz_output *out, const char *format, const char *options)
 {
-	if (is_extension(format, "cbz"))
-		return fz_new_cbz_writer_with_output(ctx, out, options);
+#if FZ_ENABLE_OCR_OUTPUT
 	if (is_extension(format, "ocr"))
 		return fz_new_pdfocr_writer_with_output(ctx, out, options);
+#endif
 #if FZ_ENABLE_PDF
 	if (is_extension(format, "pdf"))
 		return fz_new_pdf_writer_with_output(ctx, out, options);
 #endif
 
+	if (is_extension(format, "cbz"))
+		return fz_new_cbz_writer_with_output(ctx, out, options);
+
+	if (is_extension(format, "svg"))
+		return fz_new_svg_writer_with_output(ctx, out, options);
+
 	if (is_extension(format, "pcl"))
 		return fz_new_pcl_writer_with_output(ctx, out, options);
 	if (is_extension(format, "pclm"))
@@ -288,7 +296,7 @@ fz_new_document_writer_with_output(fz_context *ctx, fz_output *out, const char *
 		return fz_new_docx_writer_with_output(ctx, out, options);
 #endif
 
-	fz_throw(ctx, FZ_ERROR_GENERIC, "unknown output document format: %s", format);
+	fz_throw(ctx, FZ_ERROR_ARGUMENT, "unknown output document format: %s", format);
 }
 
 fz_document_writer *
@@ -296,12 +304,13 @@ fz_new_document_writer_with_buffer(fz_context *ctx, fz_buffer *buffer, const cha
 {
 	fz_document_writer *wri;
 	fz_output *out = fz_new_output_with_buffer(ctx, buffer);
-	fz_try(ctx)
+	fz_try(ctx) {
 		wri = fz_new_document_writer_with_output(ctx, out, format, options);
-	fz_always(ctx)
+	}
+	fz_catch(ctx) {
 		fz_drop_output(ctx, out);
-	fz_catch(ctx)
 		fz_rethrow(ctx);
+	}
 	return wri;
 }
 
@@ -334,7 +343,7 @@ fz_begin_page(fz_context *ctx, fz_document_writer *wri, fz_rect mediabox)
 	if (!wri)
 		return NULL;
 	if (wri->dev)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "called begin page without ending the previous page");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "called begin page without ending the previous page");
 	wri->dev = wri->begin_page(ctx, wri, mediabox);
 	return wri->dev;
 }
