@@ -156,7 +156,7 @@ pdf_filter_content_stream(
 	fz_try(ctx)
 	{
 		*out_buf = fz_new_buffer(ctx, 1024);
-		top = proc_buffer = pdf_new_buffer_processor(ctx, *out_buf, options->ascii);
+		top = proc_buffer = pdf_new_buffer_processor(ctx, *out_buf, options->ascii, options->newlines);
 		if (num_filters > 0)
 		{
 			for (i = num_filters - 1; i >= 0; i--)
@@ -239,7 +239,7 @@ pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page
 			in_res = page_res;
 
 		buffer = fz_new_buffer(ctx, 1024);
-		top = proc_buffer = pdf_new_buffer_processor(ctx, buffer, options->ascii);
+		top = proc_buffer = pdf_new_buffer_processor(ctx, buffer, options->ascii, options->newlines);
 		if (num_filters > 0)
 		{
 			for (i = num_filters - 1; i >= 0; i--)
@@ -253,10 +253,14 @@ pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page
 		{
 			pdf_obj *val = pdf_dict_get_val(ctx, charprocs, i);
 
-			fz_clear_buffer(ctx, buffer);
+			if (i > 0)
+			{
+				pdf_reset_processor(ctx, top);
+				fz_clear_buffer(ctx, buffer);
+			}
 			pdf_process_raw_contents(ctx, top, doc, in_res, val, NULL);
 
-			pdf_close_processor(ctx, proc_buffer);
+			pdf_close_processor(ctx, top);
 
 			if (!options->no_update)
 			{
@@ -290,7 +294,6 @@ static void
 pdf_filter_xobject(fz_context *ctx, pdf_document *doc, pdf_obj *stm, pdf_obj *page_res, pdf_filter_options *options, pdf_cycle_list *cycle_up)
 {
 	pdf_cycle_list cycle;
-	pdf_obj *struct_parents_obj;
 	int struct_parents;
 	pdf_obj *new_res = NULL;
 	fz_buffer *new_buf = NULL;
@@ -301,10 +304,7 @@ pdf_filter_xobject(fz_context *ctx, pdf_document *doc, pdf_obj *stm, pdf_obj *pa
 
 	// TODO for RJW: XObject can also be a StructParent; how do we handle that case?
 
-	struct_parents_obj = pdf_dict_get(ctx, stm, PDF_NAME(StructParents));
-	struct_parents = -1;
-	if (pdf_is_number(ctx, struct_parents_obj))
-		struct_parents = pdf_to_int(ctx, struct_parents_obj);
+	struct_parents = pdf_dict_get_int_default(ctx, stm, PDF_NAME(StructParents), -1);
 
 	old_res = pdf_dict_get(ctx, stm, PDF_NAME(Resources));
 	if (!old_res)
@@ -341,7 +341,6 @@ pdf_filter_xobject_instance(fz_context *ctx, pdf_obj *old_xobj, pdf_obj *page_re
 	pdf_obj *new_xobj;
 	pdf_obj *new_res, *old_res;
 	fz_buffer *new_buf;
-	pdf_obj *struct_parents_obj;
 	int struct_parents;
 	fz_matrix matrix;
 
@@ -352,10 +351,7 @@ pdf_filter_xobject_instance(fz_context *ctx, pdf_obj *old_xobj, pdf_obj *page_re
 	// TODO for RJW: XObject can also be a StructParent; how do we handle that case?
 	// TODO for RJW: will we run into trouble by duplicating StructParents stuff?
 
-	struct_parents_obj = pdf_dict_get(ctx, old_xobj, PDF_NAME(StructParents));
-	struct_parents = -1;
-	if (pdf_is_number(ctx, struct_parents_obj))
-		struct_parents = pdf_to_int(ctx, struct_parents_obj);
+	struct_parents = pdf_dict_get_int_default(ctx, old_xobj, PDF_NAME(StructParents), -1);
 
 	old_res = pdf_dict_get(ctx, old_xobj, PDF_NAME(Resources));
 	if (!old_res)
@@ -394,15 +390,11 @@ pdf_filter_xobject_instance(fz_context *ctx, pdf_obj *old_xobj, pdf_obj *page_re
 void pdf_filter_page_contents(fz_context *ctx, pdf_document *doc, pdf_page *page, pdf_filter_options *options)
 {
 	pdf_obj *contents, *old_res;
-	pdf_obj *struct_parents_obj;
 	pdf_obj *new_res;
 	fz_buffer *buffer;
 	int struct_parents;
 
-	struct_parents_obj = pdf_dict_get(ctx, page->obj, PDF_NAME(StructParents));
-	struct_parents = -1;
-	if (pdf_is_number(ctx, struct_parents_obj))
-		struct_parents = pdf_to_int(ctx, struct_parents_obj);
+	struct_parents = pdf_dict_get_int_default(ctx, page->obj, PDF_NAME(StructParents), -1);
 
 	contents = pdf_page_contents(ctx, page);
 	old_res = pdf_page_resources(ctx, page);
@@ -460,6 +452,8 @@ struct redact_filter_state {
 	pdf_filter_factory filter_list[2];
 	pdf_page *page;
 	pdf_annot *target; // NULL if all
+	int line_art;
+	int text;
 };
 
 static void
@@ -471,7 +465,7 @@ pdf_redact_end_page(fz_context *ctx, fz_buffer *buf, void *opaque)
 	pdf_obj *qp;
 	int i, n;
 
-	fz_append_string(ctx, buf, "0 g\n");
+	fz_append_string(ctx, buf, " 0 g\n");
 
 	for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot))
 	{
@@ -537,20 +531,24 @@ pdf_redact_text_filter(fz_context *ctx, void *opaque, int *ucsbuf, int ucslen, f
 		{
 			qp = pdf_dict_get(ctx, annot->obj, PDF_NAME(QuadPoints));
 			n = pdf_array_len(ctx, qp);
+			/* Note, we test for the intersection being a valid rectangle, NOT
+			 * a non-empty one. This is because we can have 'empty' character
+			 * boxes (say for diacritics), that while 0 width, do have a defined
+			 * position on the plane, and hence inclusion makes sense. */
 			if (n > 0)
 			{
 				for (i = 0; i < n; i += 8)
 				{
 					q = pdf_to_quad(ctx, qp, i);
 					r = fz_rect_from_quad(q);
-					if (!fz_is_empty_rect(fz_intersect_rect(bbox, r)))
+					if (fz_is_valid_rect(fz_intersect_rect(bbox, r)))
 						return 1;
 				}
 			}
 			else
 			{
 				r = pdf_dict_get_rect(ctx, annot->obj, PDF_NAME(Rect));
-				if (!fz_is_empty_rect(fz_intersect_rect(bbox, r)))
+				if (fz_is_valid_rect(fz_intersect_rect(bbox, r)))
 					return 1;
 			}
 		}
@@ -660,7 +658,7 @@ pdf_redact_image_imp(fz_context *ctx, fz_matrix ctm, fz_image *image, fz_pixmap
 }
 
 static fz_image *
-pdf_redact_image_filter_remove(fz_context *ctx, void *opaque, fz_matrix ctm, const char *name, fz_image *image)
+pdf_redact_image_filter_remove(fz_context *ctx, void *opaque, fz_matrix ctm, const char *name, fz_image *image, fz_rect clip)
 {
 	fz_pixmap *redacted = NULL;
 	struct redact_filter_state *red = opaque;
@@ -707,7 +705,57 @@ pdf_redact_image_filter_remove(fz_context *ctx, void *opaque, fz_matrix ctm, con
 }
 
 static fz_image *
-pdf_redact_image_filter_pixels(fz_context *ctx, void *opaque, fz_matrix ctm, const char *name, fz_image *image)
+pdf_redact_image_filter_remove_invisible(fz_context *ctx, void *opaque, fz_matrix ctm, const char *name, fz_image *image, fz_rect clip)
+{
+	fz_pixmap *redacted = NULL;
+	struct redact_filter_state *red = opaque;
+	pdf_page *page = red->page;
+	pdf_annot *annot;
+	pdf_obj *qp;
+	fz_rect area;
+	fz_rect r;
+	int i, n;
+
+	fz_var(redacted);
+
+	area = fz_transform_rect(fz_unit_rect, ctm);
+
+	/* Restrict the are of the image to that which can actually be seen. */
+	area = fz_intersect_rect(area, clip);
+
+	for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot))
+	{
+		if (red->target != NULL && red->target != annot)
+			continue;
+		if (pdf_dict_get(ctx, annot->obj, PDF_NAME(Subtype)) == PDF_NAME(Redact))
+		{
+			qp = pdf_dict_get(ctx, annot->obj, PDF_NAME(QuadPoints));
+			n = pdf_array_len(ctx, qp);
+			if (n > 0)
+			{
+				for (i = 0; i < n; i += 8)
+				{
+					r = fz_rect_from_quad(pdf_to_quad(ctx, qp, i));
+					r = fz_intersect_rect(r, area);
+					if (!fz_is_empty_rect(r))
+						return NULL;
+				}
+			}
+			else
+			{
+				r = pdf_dict_get_rect(ctx, annot->obj, PDF_NAME(Rect));
+				r = fz_intersect_rect(r, area);
+				if (!fz_is_empty_rect(r))
+					return NULL;
+			}
+		}
+	}
+
+	return fz_keep_image(ctx, image);
+}
+
+static fz_image *
+pdf_redact_image_filter_pixels(fz_context *ctx, void *opaque, fz_matrix ctm, const char *name, fz_image *image, fz_rect clip)
 {
 	fz_pixmap *redacted = NULL;
 	fz_pixmap *mask = NULL;
@@ -819,17 +867,21 @@ pdf_redact_image_filter_pixels(fz_context *ctx, void *opaque, fz_matrix ctm, con
 			fz_rethrow(ctx);
 		return image;
 	}
-	return fz_keep_image(ctx, image);
+	return image;
 }
 
+/* Returns 0 if area does not intersect with any of our redactions.
+ * Returns 2 if area is completely included within one of our redactions.
+ * Returns 1 otherwise. */
 static int
-rect_touches_redactions(fz_context *ctx, pdf_document *doc, pdf_page *page, fz_rect area, struct redact_filter_state *red)
+rect_touches_redactions(fz_context *ctx, fz_rect area, struct redact_filter_state *red)
 {
 	pdf_annot *annot;
 	pdf_obj *qp;
 	fz_quad q;
-	fz_rect r;
+	fz_rect r, s;
 	int i, n;
+	pdf_page *page = red->page;
 
 	for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot))
 	{
@@ -845,17 +897,25 @@ rect_touches_redactions(fz_context *ctx, pdf_document *doc, pdf_page *page, fz_r
 				{
 					q = pdf_to_quad(ctx, qp, i);
 					r = fz_rect_from_quad(q);
-					r = fz_intersect_rect(r, area);
-					if (!fz_is_empty_rect(r))
+					s = fz_intersect_rect(r, area);
+					if (!fz_is_empty_rect(s))
+					{
+						if (fz_contains_rect(r, area))
+							return 2;
 						return 1;
+					}
 				}
 			}
 			else
 			{
 				r = pdf_dict_get_rect(ctx, annot->obj, PDF_NAME(Rect));
-				r = fz_intersect_rect(r, area);
-				if (!fz_is_empty_rect(r))
+				s = fz_intersect_rect(r, area);
+				if (!fz_is_empty_rect(s))
+				{
+					if (fz_contains_rect(r, area))
+						return 2;
 					return 1;
+				}
 			}
 		}
 	}
@@ -863,14 +923,14 @@ rect_touches_redactions(fz_context *ctx, pdf_document *doc, pdf_page *page, fz_r
 }
 
 static void
-pdf_redact_page_links(fz_context *ctx, pdf_document *doc, pdf_page *page, struct redact_filter_state *red)
+pdf_redact_page_links(fz_context *ctx, struct redact_filter_state *red)
 {
 	pdf_obj *annots;
 	pdf_obj *link;
 	fz_rect area;
 	int k;
 
-	annots = pdf_dict_get(ctx, page->obj, PDF_NAME(Annots));
+	annots = pdf_dict_get(ctx, red->page->obj, PDF_NAME(Annots));
 	k = 0;
 	while (k < pdf_array_len(ctx, annots))
 	{
@@ -878,7 +938,7 @@ pdf_redact_page_links(fz_context *ctx, pdf_document *doc, pdf_page *page, struct
 		if (pdf_dict_get(ctx, link, PDF_NAME(Subtype)) == PDF_NAME(Link))
 		{
 			area = pdf_dict_get_rect(ctx, link, PDF_NAME(Rect));
-			if (rect_touches_redactions(ctx, doc, page, area, red))
+			if (rect_touches_redactions(ctx, area, red))
 			{
 				pdf_array_delete(ctx, annots, k);
 				continue;
@@ -889,31 +949,52 @@ pdf_redact_page_links(fz_context *ctx, pdf_document *doc, pdf_page *page, struct
 }
 
 static void
-pdf_redact_page_annotations(fz_context *ctx, pdf_document *doc, pdf_page *page, struct redact_filter_state *red)
+pdf_redact_page_annotations(fz_context *ctx, struct redact_filter_state *red)
 {
 	pdf_annot *annot;
 	fz_rect area;
 
 restart:
-	for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot))
+	for (annot = pdf_first_annot(ctx, red->page); annot; annot = pdf_next_annot(ctx, annot))
 	{
 		if (pdf_annot_type(ctx, annot) == PDF_ANNOT_FREE_TEXT)
 		{
 			area = pdf_dict_get_rect(ctx, pdf_annot_obj(ctx, annot), PDF_NAME(Rect));
-			if (rect_touches_redactions(ctx, doc, page, area, red))
+			if (rect_touches_redactions(ctx, area, red))
 			{
-				pdf_delete_annot(ctx, page, annot);
+				pdf_delete_annot(ctx, red->page, annot);
 				goto restart;
 			}
 		}
 	}
 }
 
+static int culler(fz_context *ctx, void *opaque, fz_rect bbox, fz_cull_type type)
+{
+	struct redact_filter_state *red = opaque;
+
+	switch (type)
+	{
+	case FZ_CULL_PATH_FILL:
+	case FZ_CULL_PATH_STROKE:
+	case FZ_CULL_PATH_FILL_STROKE:
+		if (red->line_art == PDF_REDACT_LINE_ART_REMOVE_IF_COVERED)
+			return (rect_touches_redactions(ctx, bbox, red) == 2);
+		else if (red->line_art == PDF_REDACT_LINE_ART_REMOVE_IF_TOUCHED)
+			return (rect_touches_redactions(ctx, bbox, red) != 0);
+		return 0;
+	default:
+		return 0;
+	}
+}
+
 static
 void init_redact_filter(fz_context *ctx, pdf_redact_options *redact_opts, struct redact_filter_state *red, pdf_page *page, pdf_annot *target)
 {
 	int black_boxes = redact_opts ? redact_opts->black_boxes : 0;
 	int image_method = redact_opts ? redact_opts->image_method : PDF_REDACT_IMAGE_PIXELS;
+	int line_art = redact_opts ? redact_opts->line_art : PDF_REDACT_LINE_ART_NONE;
+	int text = redact_opts ? redact_opts->text : PDF_REDACT_TEXT_REMOVE;
 
 	memset(&red->filter_opts, 0, sizeof red->filter_opts);
 	memset(&red->sanitize_opts, 0, sizeof red->sanitize_opts);
@@ -925,13 +1006,19 @@ void init_redact_filter(fz_context *ctx, pdf_redact_options *redact_opts, struct
 	red->filter_opts.filters = red->filter_list;
 	if (black_boxes)
 		red->filter_opts.complete = pdf_redact_end_page;
+	red->line_art = line_art;
+	red->text = text;
 
 	red->sanitize_opts.opaque = red;
-	red->sanitize_opts.text_filter = pdf_redact_text_filter;
+	if (text == PDF_REDACT_TEXT_REMOVE)
+		red->sanitize_opts.text_filter = pdf_redact_text_filter;
 	if (image_method == PDF_REDACT_IMAGE_PIXELS)
 		red->sanitize_opts.image_filter = pdf_redact_image_filter_pixels;
 	if (image_method == PDF_REDACT_IMAGE_REMOVE)
 		red->sanitize_opts.image_filter = pdf_redact_image_filter_remove;
+	if (image_method == PDF_REDACT_IMAGE_REMOVE_UNLESS_INVISIBLE)
+		red->sanitize_opts.image_filter = pdf_redact_image_filter_remove_invisible;
+	red->sanitize_opts.culler = culler;
 
 	red->filter_list[0].filter = pdf_new_sanitize_filter;
 	red->filter_list[0].options = &red->sanitize_opts;
@@ -943,11 +1030,12 @@ void init_redact_filter(fz_context *ctx, pdf_redact_options *redact_opts, struct
 }
 
 static int
-pdf_apply_redaction_imp(fz_context *ctx, pdf_document *doc, pdf_page *page, pdf_annot *target, pdf_redact_options *redact_opts)
+pdf_apply_redaction_imp(fz_context *ctx, pdf_page *page, pdf_annot *target, pdf_redact_options *redact_opts)
 {
 	pdf_annot *annot;
 	int has_redactions = 0;
 	struct redact_filter_state red;
+	pdf_document *doc = page->doc;
 
 	for (annot = pdf_first_annot(ctx, page); annot; annot = pdf_next_annot(ctx, annot)) {
 		if (target != NULL && target != annot)
@@ -968,8 +1056,8 @@ pdf_apply_redaction_imp(fz_context *ctx, pdf_document *doc, pdf_page *page, pdf_
 	fz_try(ctx)
 	{
 		pdf_filter_page_contents(ctx, doc, page, &red.filter_opts);
-		pdf_redact_page_links(ctx, doc, page, &red);
-		pdf_redact_page_annotations(ctx, doc, page, &red);
+		pdf_redact_page_links(ctx, &red);
+		pdf_redact_page_annotations(ctx, &red);
 
 		annot = pdf_first_annot(ctx, page);
 		while (annot)
@@ -1001,11 +1089,13 @@ pdf_apply_redaction_imp(fz_context *ctx, pdf_document *doc, pdf_page *page, pdf_
 int
 pdf_redact_page(fz_context *ctx, pdf_document *doc, pdf_page *page, pdf_redact_options *redact_opts)
 {
-	return pdf_apply_redaction_imp(ctx, doc, page, NULL, redact_opts);
+	if (page == NULL || page->doc != doc)
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Can't redact a page not from the doc");
+	return pdf_apply_redaction_imp(ctx, page, NULL, redact_opts);
 }
 
 int
 pdf_apply_redaction(fz_context *ctx, pdf_annot *annot, pdf_redact_options *redact_opts)
 {
-	return pdf_apply_redaction_imp(ctx, annot->page->doc, annot->page, annot, redact_opts);
+	return pdf_apply_redaction_imp(ctx, annot->page, annot, redact_opts);
 }
