@@ -50,6 +50,8 @@ load_icc_based(fz_context *ctx, pdf_obj *dict, int allow_alt, pdf_cycle_list *cy
 			fz_catch(ctx)
 			{
 				fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+				fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+				fz_report_error(ctx);
 				fz_warn(ctx, "ignoring broken ICC Alternate colorspace");
 			}
 		}
@@ -78,7 +80,12 @@ load_icc_based(fz_context *ctx, pdf_obj *dict, int allow_alt, pdf_cycle_list *cy
 			fz_drop_buffer(ctx, buf);
 		fz_catch(ctx)
 		{
-			fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+			if (fz_caught(ctx) == FZ_ERROR_TRYLATER || fz_caught(ctx) == FZ_ERROR_SYSTEM)
+			{
+				fz_drop_colorspace(ctx, alt);
+				fz_rethrow(ctx);
+			}
+			fz_report_error(ctx);
 			fz_warn(ctx, "ignoring broken ICC profile");
 		}
 	}
@@ -162,7 +169,7 @@ load_devicen(fz_context *ctx, pdf_obj *array, int is_devn, pdf_cycle_list *cycle
 		cs->u.separation.tint = pdf_load_function(ctx, tintobj, n, cs->u.separation.base->n);
 		if (pdf_is_array(ctx, nameobj))
 			for (i = 0; i < n; i++)
-				fz_colorspace_name_colorant(ctx, cs, i, pdf_to_name(ctx, pdf_array_get(ctx, nameobj, i)));
+				fz_colorspace_name_colorant(ctx, cs, i, pdf_array_get_name(ctx, nameobj, i));
 		else
 			fz_colorspace_name_colorant(ctx, cs, 0, pdf_to_name(ctx, nameobj));
 	}
@@ -363,7 +370,7 @@ pdf_load_colorspace_imp(fz_context *ctx, pdf_obj *obj, pdf_cycle_list *cycle_up)
 	pdf_cycle_list cycle;
 
 	if (pdf_cycle(ctx, &cycle, cycle_up, obj))
-			fz_throw(ctx, FZ_ERROR_SYNTAX, "recursive colorspace");
+		fz_throw(ctx, FZ_ERROR_SYNTAX, "recursive colorspace");
 
 	if (pdf_is_name(ctx, obj))
 	{
@@ -503,6 +510,8 @@ pdf_load_output_intent(fz_context *ctx, pdf_document *doc)
 	fz_catch(ctx)
 	{
 		fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+		fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+		fz_report_error(ctx);
 		fz_warn(ctx, "Attempt to read Output Intent failed");
 	}
 	return cs;
