@@ -48,14 +48,14 @@ pdf_close_processor(fz_context *ctx, pdf_processor *proc)
 {
 	void (*close_processor)(fz_context *ctx, pdf_processor *proc);
 
-	if (!proc)
+	if (!proc || proc->closed)
 		return;
 
+	proc->closed = 1;
 	close_processor = proc->close_processor;
 	if (!close_processor)
 		return;
 
-	proc->close_processor = NULL;
 	close_processor(ctx, proc); /* Tail recursion */
 }
 
@@ -64,7 +64,7 @@ pdf_drop_processor(fz_context *ctx, pdf_processor *proc)
 {
 	if (fz_drop_imp(ctx, proc, &proc->refs))
 	{
-		if (proc->close_processor)
+		if (!proc->closed)
 			fz_warn(ctx, "dropping unclosed PDF processor");
 		if (proc->drop_processor)
 			proc->drop_processor(ctx, proc);
@@ -72,6 +72,19 @@ pdf_drop_processor(fz_context *ctx, pdf_processor *proc)
 	}
 }
 
+void pdf_reset_processor(fz_context *ctx, pdf_processor *proc)
+{
+	if (proc == NULL)
+		return;
+
+	proc->closed = 0;
+
+	if (proc->reset_processor == NULL)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "Cannot reset PDF processor");
+
+	proc->reset_processor(ctx, proc);
+}
+
 static void
 pdf_init_csi(fz_context *ctx, pdf_csi *csi, pdf_document *doc, pdf_obj *rdb, pdf_lexbuf *buf, fz_cookie *cookie)
 {
@@ -107,8 +120,16 @@ pdf_try_load_font(fz_context *ctx, pdf_document *doc, pdf_obj *rdb, pdf_obj *fon
 	fz_catch(ctx)
 	{
 		if (fz_caught(ctx) == FZ_ERROR_TRYLATER)
+		{
+			fz_ignore_error(ctx);
 			if (cookie)
 				cookie->incomplete++;
+		}
+		else
+		{
+			fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+			fz_report_error(ctx);
+		}
 	}
 	if (desc == NULL)
 		desc = pdf_load_hail_mary_font(ctx, doc);
@@ -330,14 +351,14 @@ pdf_process_extgstate(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, pdf_ob
 				luminosity = 0;
 
 			tr = pdf_dict_get(ctx, obj, PDF_NAME(TR));
-			if (tr && !pdf_name_eq(ctx, tr, PDF_NAME(Identity)))
-				fz_warn(ctx, "ignoring transfer function");
+			if (tr && pdf_name_eq(ctx, tr, PDF_NAME(Identity)))
+				tr = NULL;
 
-			proc->op_gs_SMask(ctx, proc, xobj, softmask_bc, luminosity);
+			proc->op_gs_SMask(ctx, proc, xobj, softmask_bc, luminosity, tr);
 		}
 		else if (pdf_is_name(ctx, obj) && pdf_name_eq(ctx, obj, PDF_NAME(None)))
 		{
-			proc->op_gs_SMask(ctx, proc, NULL, NULL, 0);
+			proc->op_gs_SMask(ctx, proc, NULL, NULL, 0, NULL);
 		}
 	}
 }
@@ -350,7 +371,7 @@ pdf_process_Do(fz_context *ctx, pdf_processor *proc, pdf_csi *csi)
 	xres = pdf_dict_get(ctx, csi->rdb, PDF_NAME(XObject));
 	xobj = pdf_dict_gets(ctx, xres, csi->name);
 	if (!xobj)
-		fz_throw(ctx, FZ_ERROR_MINOR, "cannot find XObject resource '%s'", csi->name);
+		fz_throw(ctx, FZ_ERROR_SYNTAX, "cannot find XObject resource '%s'", csi->name);
 	subtype = pdf_dict_get(ctx, xobj, PDF_NAME(Subtype));
 	if (pdf_name_eq(ctx, subtype, PDF_NAME(Form)))
 	{
@@ -359,7 +380,7 @@ pdf_process_Do(fz_context *ctx, pdf_processor *proc, pdf_csi *csi)
 			subtype = st;
 	}
 	if (!pdf_is_name(ctx, subtype))
-		fz_throw(ctx, FZ_ERROR_MINOR, "no XObject subtype specified");
+		fz_throw(ctx, FZ_ERROR_SYNTAX, "no XObject subtype specified");
 
 	if (pdf_is_ocg_hidden(ctx, csi->doc, csi->rdb, proc->usage, pdf_dict_get(ctx, xobj, PDF_NAME(OC))))
 		return;
@@ -374,7 +395,10 @@ pdf_process_Do(fz_context *ctx, pdf_processor *proc, pdf_csi *csi)
 	{
 		if (proc->op_Do_image)
 		{
-			fz_image *image = pdf_load_image(ctx, csi->doc, xobj);
+			fz_image *image = NULL;
+
+			if (proc->requirements && PDF_PROCESSOR_REQUIRES_DECODED_IMAGES)
+				image = pdf_load_image(ctx, csi->doc, xobj);
 			fz_try(ctx)
 				proc->op_Do_image(ctx, proc, csi->name, image);
 			fz_always(ctx)
@@ -419,7 +443,7 @@ pdf_process_CS(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, int stroke)
 		csres = pdf_dict_get(ctx, csi->rdb, PDF_NAME(ColorSpace));
 		csobj = pdf_dict_gets(ctx, csres, csi->name);
 		if (!csobj)
-			fz_throw(ctx, FZ_ERROR_MINOR, "cannot find ColorSpace resource '%s'", csi->name);
+			fz_throw(ctx, FZ_ERROR_SYNTAX, "cannot find ColorSpace resource '%s'", csi->name);
 		if (pdf_is_array(ctx, csobj) && pdf_array_len(ctx, csobj) == 1 && pdf_name_eq(ctx, pdf_array_get(ctx, csobj, 0), PDF_NAME(Pattern)))
 		{
 			if (stroke)
@@ -449,16 +473,17 @@ pdf_process_SC(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, int stroke)
 {
 	if (csi->name[0])
 	{
-		pdf_obj *patres, *patobj, *type;
+		pdf_obj *patres, *patobj;
+		int type;
 
 		patres = pdf_dict_get(ctx, csi->rdb, PDF_NAME(Pattern));
 		patobj = pdf_dict_gets(ctx, patres, csi->name);
 		if (!patobj)
-			fz_throw(ctx, FZ_ERROR_MINOR, "cannot find Pattern resource '%s'", csi->name);
+			fz_throw(ctx, FZ_ERROR_SYNTAX, "cannot find Pattern resource '%s'", csi->name);
 
-		type = pdf_dict_get(ctx, patobj, PDF_NAME(PatternType));
+		type = pdf_dict_get_int(ctx, patobj, PDF_NAME(PatternType));
 
-		if (pdf_to_int(ctx, type) == 1)
+		if (type == 1)
 		{
 			if (proc->op_SC_pattern && proc->op_sc_pattern)
 			{
@@ -477,7 +502,7 @@ pdf_process_SC(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, int stroke)
 			}
 		}
 
-		else if (pdf_to_int(ctx, type) == 2)
+		else if (type == 2)
 		{
 			if (proc->op_SC_shade && proc->op_sc_shade)
 			{
@@ -498,7 +523,7 @@ pdf_process_SC(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, int stroke)
 
 		else
 		{
-			fz_throw(ctx, FZ_ERROR_MINOR, "unknown pattern type: %d", pdf_to_int(ctx, type));
+			fz_throw(ctx, FZ_ERROR_SYNTAX, "unknown pattern type: %d", type);
 		}
 	}
 
@@ -565,20 +590,17 @@ pdf_process_EMC(fz_context *ctx, pdf_processor *proc, pdf_csi *csi)
 static void
 pdf_process_gsave(fz_context *ctx, pdf_processor *proc, pdf_csi *csi)
 {
+	++csi->gstate;
 	if (proc->op_q)
 		proc->op_q(ctx, proc);
-	++csi->gstate;
 }
 
 static void
 pdf_process_grestore(fz_context *ctx, pdf_processor *proc, pdf_csi *csi)
 {
-	if (csi->gstate > 0)
-	{
-		if (proc->op_Q)
-			proc->op_Q(ctx, proc);
-		--csi->gstate;
-	}
+	--csi->gstate;
+	if (proc->op_Q)
+		proc->op_Q(ctx, proc);
 }
 
 static void
@@ -631,7 +653,7 @@ pdf_process_keyword(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_strea
 		if (!csi->xbalance)
 		{
 			if (is_known_bad_word(word))
-				fz_throw(ctx, FZ_ERROR_MINOR, "unknown keyword: '%s'", word);
+				fz_warn(ctx, "unknown keyword: '%s'", word);
 			else
 				fz_throw(ctx, FZ_ERROR_SYNTAX, "unknown keyword: '%s'", word);
 		}
@@ -652,7 +674,7 @@ pdf_process_keyword(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_strea
 			gsres = pdf_dict_get(ctx, csi->rdb, PDF_NAME(ExtGState));
 			gsobj = pdf_dict_gets(ctx, gsres, csi->name);
 			if (!gsobj)
-				fz_throw(ctx, FZ_ERROR_MINOR, "cannot find ExtGState resource '%s'", csi->name);
+				fz_throw(ctx, FZ_ERROR_SYNTAX, "cannot find ExtGState resource '%s'", csi->name);
 			if (proc->op_gs_begin)
 				proc->op_gs_begin(ctx, proc, csi->name, gsobj);
 			pdf_process_extgstate(ctx, proc, csi, gsobj);
@@ -802,7 +824,7 @@ pdf_process_keyword(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_strea
 			shaderes = pdf_dict_get(ctx, csi->rdb, PDF_NAME(Shading));
 			shadeobj = pdf_dict_gets(ctx, shaderes, csi->name);
 			if (!shadeobj)
-				fz_throw(ctx, FZ_ERROR_MINOR, "cannot find Shading resource '%s'", csi->name);
+				fz_throw(ctx, FZ_ERROR_SYNTAX, "cannot find Shading resource '%s'", csi->name);
 			shade = pdf_load_shading(ctx, csi->doc, shadeobj);
 			fz_try(ctx)
 				proc->op_sh(ctx, proc, csi->name, shade);
@@ -1009,6 +1031,7 @@ pdf_process_stream(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_stream
 			{
 				if (caught == FZ_ERROR_TRYLATER)
 				{
+					fz_ignore_error(ctx);
 					cookie->incomplete++;
 					tok = PDF_TOK_EOF;
 				}
@@ -1016,12 +1039,9 @@ pdf_process_stream(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_stream
 				{
 					fz_rethrow(ctx);
 				}
-				else if (caught == FZ_ERROR_MINOR)
-				{
-					cookie->errors++;
-				}
 				else if (caught == FZ_ERROR_SYNTAX)
 				{
+					fz_report_error(ctx);
 					cookie->errors++;
 					if (++syntax_errors >= MAX_SYNTAX_ERRORS)
 					{
@@ -1037,13 +1057,17 @@ pdf_process_stream(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_stream
 			else
 			{
 				if (caught == FZ_ERROR_TRYLATER)
+				{
+					fz_ignore_error(ctx);
 					tok = PDF_TOK_EOF;
+				}
 				else if (caught == FZ_ERROR_ABORT)
+				{
 					fz_rethrow(ctx);
-				else if (caught == FZ_ERROR_MINOR)
-					/* ignore minor errors */ ;
+				}
 				else if (caught == FZ_ERROR_SYNTAX)
 				{
+					fz_report_error(ctx);
 					if (++syntax_errors >= MAX_SYNTAX_ERRORS)
 					{
 						fz_warn(ctx, "too many syntax errors; ignoring rest of page");
@@ -1061,6 +1085,9 @@ pdf_process_stream(fz_context *ctx, pdf_processor *proc, pdf_csi *csi, fz_stream
 		}
 	}
 	while (tok != PDF_TOK_EOF);
+
+	if (syntax_errors > 0)
+		fz_warn(ctx, "encountered syntax errors; page may not be correct");
 }
 
 void pdf_processor_push_resources(fz_context *ctx, pdf_processor *proc, pdf_obj *res)
@@ -1225,8 +1252,7 @@ pdf_process_glyph(fz_context *ctx, pdf_processor *proc, pdf_document *doc, pdf_o
 		/* Note: Any SYNTAX errors should have been swallowed
 		 * by pdf_process_stream, but in case any escape from other
 		 * functions, recast the error type here to be safe. */
-		if (fz_caught(ctx) == FZ_ERROR_SYNTAX)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "syntax error in content stream");
+		fz_morph_error(ctx, FZ_ERROR_SYNTAX, FZ_ERROR_FORMAT);
 		fz_rethrow(ctx);
 	}
 }
