@@ -272,17 +272,17 @@ pdf_select_layer_config(fz_context *ctx, pdf_document *doc, int config)
 		if (config == 0)
 			return;
 		else
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Unknown Layer config (None known!)");
+			fz_throw(ctx, FZ_ERROR_ARGUMENT, "Unknown Layer config (None known!)");
 	}
 
 	cobj = pdf_array_get(ctx, pdf_dict_get(ctx, obj, PDF_NAME(Configs)), config);
 	if (!cobj)
 	{
 		if (config != 0)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Illegal Layer config");
+			fz_throw(ctx, FZ_ERROR_ARGUMENT, "Illegal Layer config");
 		cobj = pdf_dict_get(ctx, obj, PDF_NAME(D));
 		if (!cobj)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "No default Layer config");
+			fz_throw(ctx, FZ_ERROR_FORMAT, "No default Layer config");
 	}
 
 	pdf_drop_obj(ctx, desc->intent);
@@ -361,7 +361,7 @@ pdf_layer_config_info(fz_context *ctx, pdf_document *doc, int config_num, pdf_la
 	info->creator = NULL;
 
 	if (config_num < 0 || config_num >= desc->num_configs)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Invalid layer config number");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Invalid layer config number");
 
 	ocprops = pdf_dict_getp(ctx, pdf_trailer(ctx, doc), "Root/OCProperties");
 	if (!ocprops)
@@ -373,7 +373,7 @@ pdf_layer_config_info(fz_context *ctx, pdf_document *doc, int config_num, pdf_la
 	else if (config_num == 0)
 		obj = pdf_dict_get(ctx, ocprops, PDF_NAME(D));
 	else
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Invalid layer config number");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Invalid layer config number");
 
 	info->creator = pdf_dict_get_string(ctx, obj, PDF_NAME(Creator), NULL);
 	info->name = pdf_dict_get_string(ctx, obj, PDF_NAME(Name), NULL);
@@ -443,7 +443,7 @@ void pdf_select_layer_config_ui(fz_context *ctx, pdf_document *doc, int ui)
 	pdf_ocg_ui *entry;
 
 	if (ui < 0 || ui >= desc->num_ui_entries)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Out of range UI entry selected");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Out of range UI entry selected");
 
 	entry = &desc->ui[ui];
 	if (entry->button_flags != PDF_LAYER_UI_RADIOBOX &&
@@ -465,7 +465,7 @@ void pdf_toggle_layer_config_ui(fz_context *ctx, pdf_document *doc, int ui)
 	int selected;
 
 	if (ui < 0 || ui >= desc->num_ui_entries)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Out of range UI entry toggled");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Out of range UI entry toggled");
 
 	entry = &desc->ui[ui];
 	if (entry->button_flags != PDF_LAYER_UI_RADIOBOX &&
@@ -488,7 +488,7 @@ void pdf_deselect_layer_config_ui(fz_context *ctx, pdf_document *doc, int ui)
 	pdf_ocg_ui *entry;
 
 	if (ui < 0 || ui >= desc->num_ui_entries)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Out of range UI entry deselected");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Out of range UI entry deselected");
 
 	entry = &desc->ui[ui];
 	if (entry->button_flags != PDF_LAYER_UI_RADIOBOX &&
@@ -516,7 +516,7 @@ pdf_layer_config_ui_info(fz_context *ctx, pdf_document *doc, int ui, pdf_layer_c
 	info->type = 0;
 
 	if (ui < 0 || ui >= desc->num_ui_entries)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Out of range UI entry selected");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Out of range UI entry selected");
 
 	entry = &desc->ui[ui];
 	info->type = entry->button_flags;
@@ -551,7 +551,7 @@ ocg_intents_include(fz_context *ctx, pdf_ocg_descriptor *desc, const char *name)
 	len = pdf_array_len(ctx, desc->intent);
 	for (i=0; i < len; i++)
 	{
-		const char *intent = pdf_to_name(ctx, pdf_array_get(ctx, desc->intent, i));
+		const char *intent = pdf_array_get_name(ctx, desc->intent, i);
 		if (strcmp(intent, "All") == 0)
 			return 1;
 		if (strcmp(intent, name) == 0)
@@ -627,7 +627,7 @@ pdf_is_ocg_hidden_imp(fz_context *ctx, pdf_document *doc, pdf_obj *rdb, const ch
 			int match = 0;
 			len = pdf_array_len(ctx, obj);
 			for (i=0; i<len; i++) {
-				match |= ocg_intents_include(ctx, desc, pdf_to_name(ctx, pdf_array_get(ctx, obj, i)));
+				match |= ocg_intents_include(ctx, desc, pdf_array_get_name(ctx, obj, i));
 				if (match)
 					break;
 			}
@@ -771,7 +771,10 @@ pdf_read_ocg(fz_context *ctx, pdf_document *doc)
 	fz_catch(ctx)
 	{
 		pdf_drop_ocg(ctx, doc);
+		doc->ocg = NULL;
 		fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+		fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+		fz_report_error(ctx);
 		fz_warn(ctx, "Ignoring broken Optional Content configuration");
 		doc->ocg = fz_malloc_struct(ctx, pdf_ocg_descriptor);
 	}
