@@ -69,7 +69,7 @@ typedef struct filter_gstate
 {
 	struct filter_gstate *next;
 	int pushed;
-	int empty_clip_region;
+	fz_rect clip_rect;
 	pdf_filter_gstate pending;
 	pdf_filter_gstate sent;
 } filter_gstate;
@@ -120,7 +120,6 @@ typedef struct
 	int BT_pending;
 	int in_BT;
 	float Tm_adjust;
-	void *font_name;
 	tag_record *current_tags;
 	tag_record *pending_tags;
 	resources_stack *rstack;
@@ -179,7 +178,7 @@ create_resource_name(fz_context *ctx, pdf_sanitize_processor *p, pdf_obj *key, c
 		if (!pdf_dict_gets(ctx, res, buf))
 			return;
 	}
-	fz_throw(ctx, FZ_ERROR_GENERIC, "Cannot create unique resource name");
+	fz_throw(ctx, FZ_ERROR_LIMIT, "Cannot create unique resource name");
 }
 
 static void
@@ -193,7 +192,9 @@ filter_push(fz_context *ctx, pdf_sanitize_processor *p)
 	p->gstate = new_gstate;
 
 	pdf_keep_font(ctx, new_gstate->pending.text.font);
+	fz_keep_string(ctx, new_gstate->pending.text.fontname);
 	pdf_keep_font(ctx, new_gstate->sent.text.font);
+	fz_keep_string(ctx, new_gstate->sent.text.fontname);
 }
 
 static int
@@ -211,7 +212,9 @@ filter_pop(fz_context *ctx, pdf_sanitize_processor *p)
 			p->chain->op_Q(ctx, p->chain);
 
 	pdf_drop_font(ctx, gstate->pending.text.font);
+	fz_drop_string(ctx, gstate->pending.text.fontname);
 	pdf_drop_font(ctx, gstate->sent.text.font);
+	fz_drop_string(ctx, gstate->sent.text.fontname);
 	fz_free(ctx, gstate);
 	p->gstate = old;
 	return 0;
@@ -266,7 +269,7 @@ static void filter_flush(fz_context *ctx, pdf_sanitize_processor *p, int flush)
 	int i;
 
 	/* No point in sending anything if we're clipping it away! */
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (gstate->pushed == 0)
@@ -520,10 +523,11 @@ done_SC:
 					p->chain->op_TL(ctx, p->chain, gstate->pending.text.leading);
 			}
 			if (gstate->pending.text.font != gstate->sent.text.font ||
-				gstate->pending.text.size != gstate->sent.text.size)
+				gstate->pending.text.size != gstate->sent.text.size ||
+				gstate->pending.text.fontname != gstate->sent.text.fontname)
 			{
 				if (p->chain->op_Tf)
-					p->chain->op_Tf(ctx, p->chain, p->font_name, gstate->pending.text.font, gstate->pending.text.size);
+					p->chain->op_Tf(ctx, p->chain, fz_cstring_from_string(gstate->pending.text.fontname), gstate->pending.text.font, gstate->pending.text.size);
 			}
 			if (gstate->pending.text.render != gstate->sent.text.render)
 			{
@@ -536,8 +540,10 @@ done_SC:
 					p->chain->op_Ts(ctx, p->chain, gstate->pending.text.rise);
 			}
 			pdf_drop_font(ctx, gstate->sent.text.font);
+			fz_drop_string(ctx, gstate->sent.text.fontname);
 			gstate->sent.text = gstate->pending.text;
 			gstate->sent.text.font = pdf_keep_font(ctx, gstate->pending.text.font);
+			gstate->sent.text.fontname = fz_keep_string(ctx, gstate->pending.text.fontname);
 
 			if (p->Td_pending != 0)
 			{
@@ -842,7 +848,7 @@ adjust_for_removed_space(fz_context *ctx, pdf_sanitize_processor *p)
 {
 	filter_gstate *gstate = p->gstate;
 	float adj = gstate->pending.text.word_space;
-	adjust_text(ctx, p, adj * gstate->pending.text.scale, adj);
+	adjust_text(ctx, p, adj, adj);
 }
 
 static void
@@ -874,8 +880,7 @@ push_adjustment_to_array(fz_context *ctx, pdf_sanitize_processor *p, pdf_obj *ar
 {
 	if (p->Tm_adjust == 0)
 		return;
-	/* Negate the number here, because it will be subtracted upon use. */
-	pdf_array_push_real(ctx, arr, -p->Tm_adjust * 1000);
+	pdf_array_push_real(ctx, arr, p->Tm_adjust * 1000);
 	p->Tm_adjust = 0;
 }
 
@@ -906,7 +911,7 @@ filter_show_string(fz_context *ctx, pdf_sanitize_processor *p, unsigned char *bu
 		}
 		if (i != len)
 		{
-			adjust_text(ctx, p, p->tos.char_tx / p->gstate->pending.text.scale, p->tos.char_ty / p->gstate->pending.text.scale);
+			adjust_text(ctx, p, p->tos.char_tx / p->gstate->pending.text.scale, p->tos.char_ty);
 			i += inc;
 		}
 		if (removed_space)
@@ -963,7 +968,7 @@ filter_show_text(fz_context *ctx, pdf_sanitize_processor *p, pdf_obj *text)
 					}
 					if (j != len)
 					{
-						adjust_text(ctx, p, -p->tos.char_tx / p->gstate->pending.text.scale, -p->tos.char_ty / p->gstate->pending.text.scale);
+						adjust_text(ctx, p, p->tos.char_tx / p->gstate->pending.text.scale, p->tos.char_ty);
 						j += inc;
 					}
 					if (removed_space)
@@ -972,7 +977,7 @@ filter_show_text(fz_context *ctx, pdf_sanitize_processor *p, pdf_obj *text)
 			}
 			else
 			{
-				float tadj = pdf_to_real(ctx, item) * gstate->pending.text.size * 0.001f / p->gstate->pending.text.scale;
+				float tadj = - pdf_to_real(ctx, item) * gstate->pending.text.size * 0.001f;
 				if (fontdesc->wmode == 0)
 				{
 					adjust_text(ctx, p, tadj, 0);
@@ -1002,7 +1007,7 @@ pdf_filter_w(fz_context *ctx, pdf_processor *proc, float linewidth)
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	gstate->pending.stroke.linewidth = linewidth;
@@ -1014,7 +1019,7 @@ pdf_filter_j(fz_context *ctx, pdf_processor *proc, int linejoin)
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	gstate->pending.stroke.linejoin = linejoin;
@@ -1026,7 +1031,7 @@ pdf_filter_J(fz_context *ctx, pdf_processor *proc, int linecap)
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	gstate->pending.stroke.linecap = linecap;
@@ -1038,7 +1043,7 @@ pdf_filter_M(fz_context *ctx, pdf_processor *proc, float miterlimit)
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	gstate->pending.stroke.miterlimit = miterlimit;
@@ -1049,7 +1054,7 @@ pdf_filter_d(fz_context *ctx, pdf_processor *proc, pdf_obj *array, float phase)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1062,7 +1067,7 @@ pdf_filter_ri(fz_context *ctx, pdf_processor *proc, const char *intent)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1075,7 +1080,7 @@ pdf_filter_gs_OP(fz_context *ctx, pdf_processor *proc, int b)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1088,7 +1093,7 @@ pdf_filter_gs_op(fz_context *ctx, pdf_processor *proc, int b)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1101,7 +1106,7 @@ pdf_filter_gs_OPM(fz_context *ctx, pdf_processor *proc, int i)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1114,7 +1119,7 @@ pdf_filter_gs_UseBlackPtComp(fz_context *ctx, pdf_processor *proc, pdf_obj *name
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1127,7 +1132,7 @@ pdf_filter_i(fz_context *ctx, pdf_processor *proc, float flatness)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1140,7 +1145,7 @@ pdf_filter_gs_begin(fz_context *ctx, pdf_processor *proc, const char *name, pdf_
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, FLUSH_ALL);
@@ -1154,7 +1159,7 @@ pdf_filter_gs_BM(fz_context *ctx, pdf_processor *proc, const char *blendmode)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (p->chain->op_gs_BM)
@@ -1166,7 +1171,7 @@ pdf_filter_gs_CA(fz_context *ctx, pdf_processor *proc, float alpha)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (p->chain->op_gs_CA)
@@ -1178,7 +1183,7 @@ pdf_filter_gs_ca(fz_context *ctx, pdf_processor *proc, float alpha)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (p->chain->op_gs_ca)
@@ -1186,15 +1191,15 @@ pdf_filter_gs_ca(fz_context *ctx, pdf_processor *proc, float alpha)
 }
 
 static void
-pdf_filter_gs_SMask(fz_context *ctx, pdf_processor *proc, pdf_obj *smask, float *bc, int luminosity)
+pdf_filter_gs_SMask(fz_context *ctx, pdf_processor *proc, pdf_obj *smask, float *bc, int luminosity, pdf_obj *tr)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (p->chain->op_gs_SMask)
-		p->chain->op_gs_SMask(ctx, p->chain, smask, bc, luminosity);
+		p->chain->op_gs_SMask(ctx, p->chain, smask, bc, luminosity, tr);
 }
 
 static void
@@ -1202,7 +1207,7 @@ pdf_filter_gs_end(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (p->chain->op_gs_end)
@@ -1216,9 +1221,6 @@ pdf_filter_q(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
-		return;
-
 	filter_push(ctx, p);
 }
 
@@ -1229,7 +1231,7 @@ pdf_filter_cm(fz_context *ctx, pdf_processor *proc, float a, float b, float c, f
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 	fz_matrix ctm;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	/* If we're being given an identity matrix, don't bother sending it */
@@ -1253,14 +1255,12 @@ pdf_filter_m(fz_context *ctx, pdf_processor *proc, float x, float y)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
+	fz_moveto(ctx, p->path, x, y);
 	if (p->options->culler)
-	{
-		fz_moveto(ctx, p->path, x, y);
 		return;
-	}
 
 	filter_flush(ctx, p, FLUSH_CTM);
 	if (p->chain->op_m)
@@ -1272,14 +1272,12 @@ pdf_filter_l(fz_context *ctx, pdf_processor *proc, float x, float y)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
+	fz_lineto(ctx, p->path, x, y);
 	if (p->options->culler)
-	{
-		fz_lineto(ctx, p->path, x, y);
 		return;
-	}
 
 	filter_flush(ctx, p, FLUSH_CTM);
 	if (p->chain->op_l)
@@ -1291,14 +1289,12 @@ pdf_filter_c(fz_context *ctx, pdf_processor *proc, float x1, float y1, float x2,
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
+	fz_curveto(ctx, p->path, x1, y1, x2, y2, x3, y3);
 	if (p->options->culler)
-	{
-		fz_curveto(ctx, p->path, x1, y1, x2, y2, x3, y3);
 		return;
-	}
 
 	filter_flush(ctx, p, FLUSH_CTM);
 	if (p->chain->op_c)
@@ -1310,14 +1306,12 @@ pdf_filter_v(fz_context *ctx, pdf_processor *proc, float x2, float y2, float x3,
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
+	fz_curvetov(ctx, p->path, x2, y2, x3, y3);
 	if (p->options->culler)
-	{
-		fz_curvetov(ctx, p->path, x2, y2, x3, y3);
 		return;
-	}
 
 	filter_flush(ctx, p, FLUSH_CTM);
 	if (p->chain->op_v)
@@ -1329,14 +1323,12 @@ pdf_filter_y(fz_context *ctx, pdf_processor *proc, float x1, float y1, float x3,
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
+	fz_curvetoy(ctx, p->path, x1, y1, x3, y3);
 	if (p->options->culler)
-	{
-		fz_curvetoy(ctx, p->path, x1, y1, x3, y3);
 		return;
-	}
 
 	filter_flush(ctx, p, FLUSH_CTM);
 	if (p->chain->op_y)
@@ -1348,14 +1340,12 @@ pdf_filter_h(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
+	fz_closepath(ctx, p->path);
 	if (p->options->culler)
-	{
-		fz_closepath(ctx, p->path);
 		return;
-	}
 
 	filter_flush(ctx, p, FLUSH_CTM);
 	if (p->chain->op_h)
@@ -1367,14 +1357,12 @@ pdf_filter_re(fz_context *ctx, pdf_processor *proc, float x, float y, float w, f
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
+	fz_rectto(ctx, p->path, x, y, x+w, y+h);
 	if (p->options->culler)
-	{
-		fz_rectto(ctx, p->path, x, y, x+w, y+h);
 		return;
-	}
 
 	filter_flush(ctx, p, FLUSH_CTM);
 	if (p->chain->op_re)
@@ -1442,7 +1430,7 @@ cull_replay_rectto(fz_context *ctx, void *arg, float x1, float y1, float x2, flo
 	pdf_sanitize_processor *p = (pdf_sanitize_processor *)arg;
 
 	if (p->chain->op_re)
-		p->chain->op_re(ctx, p->chain, x1, y1, x2, y2);
+		p->chain->op_re(ctx, p->chain, x1, y1, x2-x1, y2-y1);
 }
 
 typedef struct
@@ -1469,13 +1457,15 @@ end_segment(fz_context *ctx, segmenter_data_t *sd)
 		cull_replay_rectto
 	};
 	fz_rect r;
+	const fz_stroke_state *st;
 
 	if (sd->segment == NULL)
 		return;
 
-	r = fz_bound_path(ctx, sd->segment, (sd->type == FZ_CULL_PATH_STROKE || sd->type == FZ_CULL_PATH_FILL_STROKE) ? &sd->sstate : NULL, sd->ctm);
+	st = (sd->type == FZ_CULL_PATH_STROKE || sd->type == FZ_CULL_PATH_FILL_STROKE) ? &sd->sstate : NULL;
+	r = fz_bound_path(ctx, sd->segment, st, sd->ctm);
 
-	if (sd->p->options->culler(ctx, sd->p->options->opaque, r, sd->type))
+	if (sd->p->options->culler && sd->p->options->culler(ctx, sd->p->options->opaque, r, sd->type))
 	{
 		/* This segment can be skipped */
 	}
@@ -1566,7 +1556,12 @@ cull_path(fz_context *ctx, pdf_sanitize_processor *p, int type)
 	};
 
 	if (p->options->culler == NULL)
+	{
+		fz_drop_path(ctx, p->path);
+		p->path = NULL;
+		p->path = fz_new_path(ctx);
 		return 0;
+	}
 
 	sd.ctm = fz_concat(p->gstate->pending.ctm, p->gstate->sent.ctm);
 	sd.ctm = fz_concat(sd.ctm, p->transform);
@@ -1611,7 +1606,7 @@ pdf_filter_S(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_STROKE))
@@ -1627,7 +1622,7 @@ pdf_filter_s(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_STROKE))
@@ -1643,15 +1638,13 @@ pdf_filter_F(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_FILL))
 		return;
 
 	filter_flush(ctx, p, FLUSH_FILL);
-	if (p->gstate->empty_clip_region)
-		return;
 	if (p->chain->op_F)
 		p->chain->op_F(ctx, p->chain);
 }
@@ -1661,7 +1654,7 @@ pdf_filter_f(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_FILL))
@@ -1677,7 +1670,7 @@ pdf_filter_fstar(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_FILL))
@@ -1693,7 +1686,7 @@ pdf_filter_B(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_FILL_STROKE))
@@ -1709,7 +1702,7 @@ pdf_filter_Bstar(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_FILL_STROKE))
@@ -1725,7 +1718,7 @@ pdf_filter_b(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_FILL_STROKE))
@@ -1741,7 +1734,7 @@ pdf_filter_bstar(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_PATH_FILL_STROKE))
@@ -1757,15 +1750,12 @@ pdf_filter_n(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
-	if (p->options->culler)
-	{
-		fz_drop_path(ctx, p->path);
-		p->path = NULL;
-		p->path = fz_new_path(ctx);
-	}
+	fz_drop_path(ctx, p->path);
+	p->path = NULL;
+	p->path = fz_new_path(ctx);
 
 	filter_flush(ctx, p, FLUSH_CTM);
 	if (p->chain->op_n)
@@ -1778,8 +1768,15 @@ static void
 pdf_filter_W(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
+	fz_rect r;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
+		return;
+
+	r = fz_bound_path(ctx, p->path, NULL, p->gstate->pending.ctm);
+	p->gstate->clip_rect = fz_intersect_rect(p->gstate->clip_rect, r);
+
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_CLIP_PATH))
@@ -1794,8 +1791,15 @@ static void
 pdf_filter_Wstar(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
+	fz_rect r;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
+		return;
+
+	r = fz_bound_path(ctx, p->path, NULL, p->gstate->pending.ctm);
+	p->gstate->clip_rect = fz_intersect_rect(p->gstate->clip_rect, r);
+
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (cull_path(ctx, p, FZ_CULL_CLIP_PATH))
@@ -1813,7 +1817,7 @@ pdf_filter_BT(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1831,7 +1835,7 @@ pdf_filter_ET(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (!p->BT_pending)
@@ -1849,7 +1853,7 @@ pdf_filter_ET(fz_context *ctx, pdf_processor *proc)
 		 * clip path at all. We actually want it to be an empty clip path,
 		 * but we can't easily achieve that. So we have to do the clipping
 		 * ourselves. */
-		p->gstate->empty_clip_region = 1;
+		p->gstate->clip_rect = fz_empty_rect;
 	}
 	p->BT_pending = 0;
 	if (p->options->after_text_object)
@@ -1883,7 +1887,7 @@ pdf_filter_Tc(fz_context *ctx, pdf_processor *proc, float charspace)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1895,7 +1899,7 @@ pdf_filter_Tw(fz_context *ctx, pdf_processor *proc, float wordspace)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1909,7 +1913,7 @@ pdf_filter_Tz(fz_context *ctx, pdf_processor *proc, float scale)
 	 * in the gstate. */
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1921,7 +1925,7 @@ pdf_filter_TL(fz_context *ctx, pdf_processor *proc, float leading)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1933,13 +1937,13 @@ pdf_filter_Tf(fz_context *ctx, pdf_processor *proc, const char *name, pdf_font_d
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
-	fz_free(ctx, p->font_name);
-	p->font_name = NULL;
-	p->font_name = name ? fz_strdup(ctx, name) : NULL;
+	fz_drop_string(ctx, p->gstate->pending.text.fontname);
+	p->gstate->pending.text.fontname = NULL;
+	p->gstate->pending.text.fontname = name ? fz_new_string(ctx, name) : NULL;
 	pdf_drop_font(ctx, p->gstate->pending.text.font);
 	p->gstate->pending.text.font = pdf_keep_font(ctx, font);
 	p->gstate->pending.text.size = size;
@@ -1951,7 +1955,7 @@ pdf_filter_Tr(fz_context *ctx, pdf_processor *proc, int render)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1963,7 +1967,7 @@ pdf_filter_Ts(fz_context *ctx, pdf_processor *proc, float rise)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -1977,7 +1981,7 @@ pdf_filter_Td(fz_context *ctx, pdf_processor *proc, float tx, float ty)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	p->Tm_adjust = 0;
@@ -1996,7 +2000,7 @@ pdf_filter_TD(fz_context *ctx, pdf_processor *proc, float tx, float ty)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	p->gstate->pending.text.leading = -ty;
@@ -2008,7 +2012,7 @@ pdf_filter_Tm(fz_context *ctx, pdf_processor *proc, float a, float b, float c, f
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	pdf_tos_set_matrix(&p->tos, a, b, c, d, e, f);
@@ -2022,7 +2026,7 @@ pdf_filter_Tstar(fz_context *ctx, pdf_processor *proc)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	p->Tm_adjust = 0;
@@ -2042,7 +2046,7 @@ pdf_filter_TJ(fz_context *ctx, pdf_processor *proc, pdf_obj *array)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_show_text(ctx, p, array);
@@ -2053,7 +2057,7 @@ pdf_filter_Tj(fz_context *ctx, pdf_processor *proc, char *str, size_t len)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_show_string(ctx, p, (unsigned char *)str, len);
@@ -2065,7 +2069,7 @@ pdf_filter_squote(fz_context *ctx, pdf_processor *proc, char *str, size_t len)
 	/* Note, we convert all T' operators to (maybe) a T* and a Tj */
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	/* need to flush text state and clear adjustment before we emit a newline */
@@ -2088,7 +2092,7 @@ pdf_filter_dquote(fz_context *ctx, pdf_processor *proc, float aw, float ac, char
 	 * (maybe) Tc, (maybe) Tw and a Tj. */
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	/* need to flush text state and clear adjustment before we emit a newline */
@@ -2113,7 +2117,7 @@ pdf_filter_d0(fz_context *ctx, pdf_processor *proc, float wx, float wy)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -2126,7 +2130,7 @@ pdf_filter_d1(fz_context *ctx, pdf_processor *proc, float wx, float wy, float ll
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -2182,7 +2186,7 @@ pdf_filter_CS(fz_context *ctx, pdf_processor *proc, const char *name, fz_colorsp
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	fz_strlcpy(gstate->pending.CS.name, name, sizeof gstate->pending.CS.name);
@@ -2197,7 +2201,7 @@ pdf_filter_cs(fz_context *ctx, pdf_processor *proc, const char *name, fz_colorsp
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	fz_strlcpy(gstate->pending.cs.name, name, sizeof gstate->pending.cs.name);
@@ -2213,7 +2217,7 @@ pdf_filter_SC_pattern(fz_context *ctx, pdf_processor *proc, const char *name, pd
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 	int i;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	fz_strlcpy(gstate->pending.SC.name, name, sizeof gstate->pending.SC.name);
@@ -2232,7 +2236,7 @@ pdf_filter_sc_pattern(fz_context *ctx, pdf_processor *proc, const char *name, pd
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 	int i;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	fz_strlcpy(gstate->pending.sc.name, name, sizeof gstate->pending.sc.name);
@@ -2250,7 +2254,7 @@ pdf_filter_SC_shade(fz_context *ctx, pdf_processor *proc, const char *name, fz_s
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	fz_strlcpy(gstate->pending.SC.name, name, sizeof gstate->pending.SC.name);
@@ -2266,7 +2270,7 @@ pdf_filter_sc_shade(fz_context *ctx, pdf_processor *proc, const char *name, fz_s
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	fz_strlcpy(gstate->pending.sc.name, name, sizeof gstate->pending.sc.name);
@@ -2283,7 +2287,7 @@ pdf_filter_SC_color(fz_context *ctx, pdf_processor *proc, int n, float *color)
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 	int i;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	gstate->pending.SC.name[0] = 0;
@@ -2301,7 +2305,7 @@ pdf_filter_sc_color(fz_context *ctx, pdf_processor *proc, int n, float *color)
 	filter_gstate *gstate = gstate_to_update(ctx, p);
 	int i;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	gstate->pending.sc.name[0] = 0;
@@ -2318,7 +2322,7 @@ pdf_filter_G(fz_context *ctx, pdf_processor *proc, float g)
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	float color[1] = { g };
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	pdf_filter_CS(ctx, proc, "DeviceGray", fz_device_gray(ctx));
@@ -2331,7 +2335,7 @@ pdf_filter_g(fz_context *ctx, pdf_processor *proc, float g)
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	float color[1] = { g };
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	pdf_filter_cs(ctx, proc, "DeviceGray", fz_device_gray(ctx));
@@ -2344,7 +2348,7 @@ pdf_filter_RG(fz_context *ctx, pdf_processor *proc, float r, float g, float b)
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	float color[3] = { r, g, b };
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	pdf_filter_CS(ctx, proc, "DeviceRGB", fz_device_rgb(ctx));
@@ -2357,7 +2361,7 @@ pdf_filter_rg(fz_context *ctx, pdf_processor *proc, float r, float g, float b)
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	float color[3] = { r, g, b };
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	pdf_filter_cs(ctx, proc, "DeviceRGB", fz_device_rgb(ctx));
@@ -2370,7 +2374,7 @@ pdf_filter_K(fz_context *ctx, pdf_processor *proc, float c, float m, float y, fl
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	float color[4] = { c, m, y, k };
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	pdf_filter_CS(ctx, proc, "DeviceCMYK", fz_device_cmyk(ctx));
@@ -2383,7 +2387,7 @@ pdf_filter_k(fz_context *ctx, pdf_processor *proc, float c, float m, float y, fl
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	float color[4] = { c, m, y, k };
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	pdf_filter_cs(ctx, proc, "DeviceCMYK", fz_device_cmyk(ctx));
@@ -2397,7 +2401,7 @@ pdf_filter_BI(fz_context *ctx, pdf_processor *proc, fz_image *image, const char
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (p->options->culler)
@@ -2417,7 +2421,7 @@ pdf_filter_BI(fz_context *ctx, pdf_processor *proc, fz_image *image, const char
 		if (p->options->image_filter)
 		{
 			fz_matrix ctm = fz_concat(p->gstate->sent.ctm, p->transform);
-			image = p->options->image_filter(ctx, p->options->opaque, ctm, "<inline>", image);
+			image = p->options->image_filter(ctx, p->options->opaque, ctm, "<inline>", image, p->gstate->clip_rect);
 			if (image)
 			{
 				fz_try(ctx)
@@ -2444,7 +2448,7 @@ pdf_filter_sh(fz_context *ctx, pdf_processor *proc, const char *name, fz_shade *
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (p->options->culler)
@@ -2471,7 +2475,7 @@ pdf_filter_Do_image(fz_context *ctx, pdf_processor *proc, const char *name, fz_i
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	fz_image *new_image;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	if (p->options->culler)
@@ -2491,7 +2495,7 @@ pdf_filter_Do_image(fz_context *ctx, pdf_processor *proc, const char *name, fz_i
 		if (p->options->image_filter)
 		{
 			fz_matrix ctm = fz_concat(p->gstate->sent.ctm, p->transform);
-			new_image = p->options->image_filter(ctx, p->options->opaque, ctm, name, image);
+			new_image = p->options->image_filter(ctx, p->options->opaque, ctm, name, image, p->gstate->clip_rect);
 		}
 		else
 		{
@@ -2544,7 +2548,7 @@ pdf_filter_Do_form(fz_context *ctx, pdf_processor *proc, const char *name, pdf_o
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 	fz_matrix transform;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, FLUSH_ALL);
@@ -2583,7 +2587,7 @@ pdf_filter_MP(fz_context *ctx, pdf_processor *proc, const char *tag)
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -2596,7 +2600,7 @@ pdf_filter_DP(fz_context *ctx, pdf_processor *proc, const char *tag, pdf_obj *ra
 {
 	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
 
-	if (p->gstate->empty_clip_region)
+	if (fz_is_empty_rect(p->gstate->clip_rect))
 		return;
 
 	filter_flush(ctx, p, 0);
@@ -2758,7 +2762,9 @@ pdf_drop_sanitize_processor(fz_context *ctx, pdf_processor *proc)
 	{
 		filter_gstate *next = gs->next;
 		pdf_drop_font(ctx, gs->pending.text.font);
+		fz_drop_string(ctx, gs->pending.text.fontname);
 		pdf_drop_font(ctx, gs->sent.text.font);
+		fz_drop_string(ctx, gs->sent.text.fontname);
 		fz_free(ctx, gs);
 		gs = next;
 	}
@@ -2768,7 +2774,6 @@ pdf_drop_sanitize_processor(fz_context *ctx, pdf_processor *proc)
 		pop_tag(ctx, p, &p->current_tags);
 	pdf_drop_obj(ctx, p->structarray);
 	pdf_drop_document(ctx, p->doc);
-	fz_free(ctx, p->font_name);
 
 	fz_drop_path(ctx, p->path);
 
@@ -2820,6 +2825,14 @@ pdf_sanitize_pop_resources(fz_context *ctx, pdf_processor *proc)
 	return pdf_processor_pop_resources(ctx, p->chain);
 }
 
+static void
+pdf_reset_sanitize_processor(fz_context *ctx, pdf_processor *proc)
+{
+	pdf_sanitize_processor *p = (pdf_sanitize_processor*)proc;
+
+	pdf_reset_processor(ctx, p->chain);
+}
+
 pdf_processor *
 pdf_new_sanitize_filter(
 	fz_context *ctx,
@@ -2835,6 +2848,7 @@ pdf_new_sanitize_filter(
 
 	proc->super.close_processor = pdf_close_sanitize_processor;
 	proc->super.drop_processor = pdf_drop_sanitize_processor;
+	proc->super.reset_processor = pdf_reset_sanitize_processor;
 
 	proc->super.push_resources = pdf_sanitize_push_resources;
 	proc->super.pop_resources = pdf_sanitize_pop_resources;
@@ -2969,8 +2983,7 @@ pdf_new_sanitize_filter(
 
 	fz_try(ctx)
 	{
-		if (proc->options->culler)
-			proc->path = fz_new_path(ctx);
+		proc->path = fz_new_path(ctx);
 		proc->gstate = fz_malloc_struct(ctx, filter_gstate);
 		proc->gstate->pending.ctm = fz_identity;
 		proc->gstate->sent.ctm = fz_identity;
@@ -2982,6 +2995,7 @@ pdf_new_sanitize_filter(
 		proc->gstate->sent.text.size = -1;
 		proc->gstate->sent.stroke.linewidth = 1;
 		proc->gstate->sent.stroke.miterlimit = 10;
+		proc->gstate->clip_rect = fz_infinite_rect;
 	}
 	fz_catch(ctx)
 	{
@@ -2989,5 +3003,7 @@ pdf_new_sanitize_filter(
 		fz_rethrow(ctx);
 	}
 
+	proc->super.requirements = proc->chain->requirements;
+
 	return (pdf_processor*)proc;
 }
