@@ -81,6 +81,8 @@ pdf_xobject_colorspace(fz_context *ctx, pdf_obj *xobj)
 			fz_catch(ctx)
 			{
 				fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+				fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+				fz_report_error(ctx);
 				fz_warn(ctx, "Ignoring XObject blending colorspace.");
 			}
 			if (!fz_is_valid_blend_colorspace(ctx, colorspace))
